# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2019, 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

__all__ = ['FlashPureVLS']

from thermo.flash.flash_base import Flash
from fluids.numerics import (
    numpy as np,
    secant,
    brenth,
    newton,
    linspace,
    assert_close,
    UnconvergedError,
    NoSolutionError
)
from chemicals.utils import rho_to_Vm, Vm_to_rho
from chemicals.exceptions import PhaseExistenceImpossible
from thermo.bulk import default_settings
from thermo.eos_mix import IGMIX
from thermo.coolprop import CPiP_min
from thermo.phases import coolprop_phase
from thermo.phases.coolprop_phase import (
    CPPQ_INPUTS,
    CPQT_INPUTS, 
    CPunknown, 
    CPiDmolar,
)
from chemicals.iapws import (
    iapws95_Psat, 
    iapws95_Tsat, 
    iapws95_rhog_sat, 
    iapws95_rhol_sat, 
    iapws95_Tc, 
    iapws95_Pc, 
    iapws95_MW, 
    iapws95_T
)
from thermo.phases import (
    Phase,
    CEOSGas, 
    CEOSLiquid, 
    IdealGas, 
    CoolPropGas, 
    CoolPropLiquid,
    IAPWS95Gas,
    IAPWS95Liquid,
    GibbsExcessLiquid,
    DryAirLemmon
)
from thermo.flash.flash_utils import (
    TVF_pure_secant,
    PVF_pure_newton,
    TSF_pure_newton,
    PSF_pure_newton,
    solve_PTV_HSGUA_1P
)


class FlashPureVLS(Flash):
    r'''Class for performing flash calculations on pure-component systems.
    This class is subtantially more robust than using multicomponent algorithms
    on pure species. It is also faster. All parameters are also attributes.

    The minimum information that is needed in addition to the :obj:`Phase`
    objects is:

    * MW
    * Vapor pressure curve if including liquids
    * Sublimation pressure curve if including solids
    * Functioning enthalpy models for each phase

    Parameters
    ----------
    constants : :obj:`ChemicalConstantsPackage <thermo.chemical_package.ChemicalConstantsPackage>` object
        Package of chemical constants; these are used as boundaries at times,
        initial guesses other times, and in all cases these properties are
        accessible as attributes of the resulting
        :obj:`EquilibriumState <thermo.equilibrium.EquilibriumState>` object, [-]
    correlations : :obj:`PropertyCorrelationsPackage <thermo.chemical_package.PropertyCorrelationsPackage>`
        Package of chemical T-dependent properties; these are used as boundaries at times,
        for initial guesses other times, and in all cases these properties are
        accessible as attributes of the resulting
        :obj:`EquilibriumState <thermo.equilibrium.EquilibriumState>` object, [-]
    gas : :obj:`Phase <thermo.phases.Phase>` object
        A single phase which can represent the gas phase, [-]
    liquids : list[:obj:`Phase <thermo.phases.Phase>`]
        A list of phases for representing the liquid phase; normally only one
        liquid phase is present for a pure-component system, but multiple
        liquids are allowed for the really weird cases like having both
        parahydrogen and orthohydrogen. The liquid phase which calculates a
        lower Gibbs free energy is always used. [-]
    solids : list[:obj:`Phase <thermo.phases.Phase>`]
        A list of phases for representing the solid phase; it is very common
        for multiple solid forms of a compound to exist. For water ice, the
        list is very long - normally ice is in phase Ih but other phases are Ic,
        II, III, IV, V, VI, VII, VIII, IX, X, XI, XII, XIII, XIV, XV, XVI,
        Square ice, and Amorphous ice. It is less common for there to be
        published, reliable, thermodynamic models for these different phases;
        for water there is the IAPWS-06 model for Ih, and another model
        `here <https://aip.scitation.org/doi/10.1063/1.1931662>`_
        for phases Ih, Ic, II, III, IV, V, VI, IX, XI, XII. [-]
    settings : :obj:`BulkSettings <thermo.bulk.BulkSettings>` object
        Object containing settings for calculating bulk and transport
        properties, [-]

    Attributes
    ----------
    VL_IG_hack : bool
        Whether or not to trust the saturation curve of the liquid phase;
        applied automatically to the
        :obj:`GibbsExcessLiquid <thermo.phases.GibbsExcessLiquid>`
        phase if there is a single liquid only, [-]
    VL_EOS_hacks : bool
        Whether or not to trust the saturation curve of the EOS liquid phase;
        applied automatically to the
        :obj:`CEOSLiquid <thermo.phases.CEOSLiquid>`
        phase if there is a single liquid only, [-]
    TPV_HSGUA_guess_maxiter : int
        Maximum number of iterations to try when converging a shortcut model
        for flashes with one (`T`, `P`, `V`) spec and one (`H`, `S`, `G`, `U`,
        `A`) spec, [-]
    TPV_HSGUA_guess_xtol : float
        Convergence tolerance in the iteration variable when converging a
        shortcut model for flashes with one (`T`, `P`, `V`) spec and one (`H`,
        `S`, `G`, `U`, `A`) spec, [-]
    TPV_HSGUA_maxiter : int
        Maximum number of iterations to try when converging a flashes with one
        (`T`, `P`, `V`) spec and one (`H`, `S`, `G`, `U`, `A`) spec; this is
        on a per-phase basis, so if there is a liquid and a gas phase, the
        maximum number of iterations that could end up being tried would be
        twice this, [-]
    TPV_HSGUA_xtol : float
        Convergence tolerance in the iteration variable dimension when
        converging a flash with one (`T`, `P`, `V`) spec and one (`H`, `S`,
        `G`, `U`, `A`) spec, [-]
    TVF_maxiter : int
        Maximum number of iterations to try when converging a flashes with a
        temperature and vapor fraction specification, [-]
    TVF_xtol : float
        Convergence tolerance in the temperature dimension when converging a
        flashes with a temperature and vapor fraction specification, [-]
    PVF_maxiter : int
        Maximum number of iterations to try when converging a flashes with a
        pressure and vapor fraction specification, [-]
    PVF_xtol : float
        Convergence tolerance in the pressure dimension when converging a
        flashes with a pressure and vapor fraction specification, [-]
    TSF_maxiter : int
        Maximum number of iterations to try when converging a flashes with a
        temperature and solid fraction specification, [-]
    TSF_xtol : float
        Convergence tolerance in the temperature dimension when converging a
        flashes with a temperature and solid fraction specification, [-]
    PSF_maxiter : int
        Maximum number of iterations to try when converging a flashes with a
        pressure and solid fraction specification, [-]
    PSF_xtol : float
        Convergence tolerance in the pressure dimension when converging a
        flashes with a pressure and solid fraction specification, [-]


    Notes
    -----
    The algorithms in this object are mostly from [1]_ and [2]_. They all
    boil down to newton methods with analytical derivatives. The phase with
    the lowest Gibbs energy is the most stable if there are multiple
    solutions.

    Phase input combinations which have specific simplifying assumptions
    (and thus more speed) are:

    * a :obj:`CEOSLiquid <thermo.phases.CEOSLiquid>` and a :obj:`CEOSGas <thermo.phases.CEOSGas>` with the same (consistent) parameters
    * a :obj:`CEOSGas <thermo.phases.CEOSGas>` with the :obj:`IGMIX <thermo.eos_mix.IGMIX>` eos and a :obj:`GibbsExcessLiquid <thermo.phases.GibbsExcessLiquid>`
    * a :obj:`IAPWS95Liquid <thermo.phases.IAPWS95Liquid>` and a :obj:`IAPWS95Gas <thermo.phases.IAPWS95Gas>`
    * a :obj:`CoolPropLiquid <thermo.phases.CoolPropLiquid>` and a :obj:`CoolPropGas <thermo.phases.CoolPropGas>`

    Additional information that can be provided in the
    :obj:`ChemicalConstantsPackage <thermo.chemical_package.ChemicalConstantsPackage>`
    object and :obj:`PropertyCorrelationsPackage <thermo.chemical_package.PropertyCorrelationsPackage>`
    object that may help convergence is:

    * `Tc`, `Pc`, `omega`, `Tb`, and `atoms`
    * Gas heat capacity correlations
    * Liquid molar volume correlations
    * Heat of vaporization correlations

    Examples
    --------

    Create all the necessary objects using all of the default parameters for
    decane and do a flash at 300 K and 1 bar:

    >>> from thermo import ChemicalConstantsPackage, PRMIX, CEOSLiquid, CEOSGas, FlashPureVLS
    >>> constants, correlations = ChemicalConstantsPackage.from_IDs(['decane'])
    >>> eos_kwargs = dict(Tcs=constants.Tcs, Pcs=constants.Pcs, omegas=constants.omegas)
    >>> liquid = CEOSLiquid(PRMIX, HeatCapacityGases=correlations.HeatCapacityGases, eos_kwargs=eos_kwargs)
    >>> gas = CEOSGas(PRMIX, HeatCapacityGases=correlations.HeatCapacityGases, eos_kwargs=eos_kwargs)
    >>> flasher = FlashPureVLS(constants, correlations, gas=gas, liquids=[liquid], solids=[])
    >>> print(flasher.flash(T=300, P=1e5))
    <EquilibriumState, T=300.0000, P=100000.0000, zs=[1.0], betas=[1.0], phases=[<CEOSLiquid, T=300 K, P=100000 Pa>]>

    Working with steam:

    >>> from thermo import FlashPureVLS, IAPWS95Liquid, IAPWS95Gas, iapws_constants, iapws_correlations
    >>> liquid = IAPWS95Liquid(T=300, P=1e5, zs=[1])
    >>> gas = IAPWS95Gas(T=300, P=1e5, zs=[1])
    >>> flasher = FlashPureVLS(iapws_constants, iapws_correlations, gas, [liquid], [])
    >>> PT = flasher.flash(T=800.0, P=1e7)
    >>> PT.rho_mass()
    29.1071839176
    >>> print(flasher.flash(T=600, VF=.5))
    <EquilibriumState, T=600.0000, P=12344824.3572, zs=[1.0], betas=[0.5, 0.5], phases=[<IAPWS95Gas, T=600 K, P=1.23448e+07 Pa>, <IAPWS95Liquid, T=600 K, P=1.23448e+07 Pa>]>
    >>> print(flasher.flash(T=600.0, H=50802))
    <EquilibriumState, T=600.0000, P=10000469.1288, zs=[1.0], betas=[1.0], phases=[<IAPWS95Gas, T=600 K, P=1.00005e+07 Pa>]>
    >>> print(flasher.flash(P=1e7, S=104.))
    <EquilibriumState, T=599.6790, P=10000000.0000, zs=[1.0], betas=[1.0], phases=[<IAPWS95Gas, T=599.679 K, P=1e+07 Pa>]>
    >>> print(flasher.flash(V=.00061, U=55850))
    <EquilibriumState, T=800.5922, P=10144789.0899, zs=[1.0], betas=[1.0], phases=[<IAPWS95Gas, T=800.592 K, P=1.01448e+07 Pa>]>

    References
    ----------
    .. [1] Poling, Bruce E., John M. Prausnitz, and John P. O’Connell. The
       Properties of Gases and Liquids. 5th edition. New York: McGraw-Hill
       Professional, 2000.
    .. [2] Gmehling, Jürgen, Michael Kleiber, Bärbel Kolbe, and Jürgen Rarey.
       Chemical Thermodynamics for Process Simulation. John Wiley & Sons, 2019.
    '''
    VF_interpolators_built = False
    N = 1
    VL_EOS_hacks = True
    VL_IG_hack = True

    TPV_HSGUA_guess_maxiter = 50
    TPV_HSGUA_guess_xtol = 1e-7
    TPV_HSGUA_maxiter = 80
    TPV_HSGUA_xtol = 1e-10

    TVF_maxiter = 200
    TVF_xtol = 1e-10

    PVF_maxiter = 200
    PVF_xtol = 1e-10

    TSF_maxiter = 200
    TSF_xtol = 1e-10

    PSF_maxiter = 200
    PSF_xtol = 1e-10

    def __repr__(self):
        return "FlashPureVLS(gas=%s, liquids=%s, solids=%s)" %(self.gas, self.liquids, self.solids)
    def __init__(self, constants, correlations, gas, liquids, solids,
                 settings=default_settings):
        # These attributes are all that needs to be stored, then call _finish_initialization
        self.constants = constants
        self.correlations = correlations
        self.solids = solids
        self.liquids = liquids
        self.gas = gas
        self.settings = settings

        self._finish_initialization()

    def _finish_initialization(self):
        solids = self.solids
        liquids = self.liquids
        gas = self.gas


        self.gas_count = 1 if gas is not None else 0
        self.liquid_count = len(liquids)
        self.liquid = liquids[0] if len(liquids) else None
        self.solid_count = len(solids)
        
        self.supports_VF_flash = self.gas_count != 0 and self.liquid_count != 0
        self.supports_SF_flash = (self.gas_count != 0 or self.liquid_count != 0) and self.solid_count != 0

        self.skip_solids = not bool(solids)

        self.phase_count = self.gas_count + self.liquid_count + self.solid_count

        if gas is not None:
            phases = [gas] + liquids + solids

        else:
            phases = liquids + solids
        self.phases = phases


        for i, l in enumerate(self.liquids):
            setattr(self, 'liquid' + str(i), l)
        for i, s in enumerate(self.solids):
            setattr(self, 'solid' + str(i), s)

        self.VL_only = self.phase_count == 2 and self.liquid_count == 1 and self.gas is not None
        self.VL_only_CEOSs = (self.VL_only 
                              and gas 
                              and liquids 
                              and isinstance(self.liquids[0], CEOSLiquid) 
                              and isinstance(self.gas, CEOSGas))

        self.VL_only_IAPWS95 = (len(liquids) == 1
                                and (isinstance(liquids[0], IAPWS95Liquid)
                                     or liquids[0].__class__.__name__ == 'IAPWS95Liquid')
                                 and (isinstance(gas, IAPWS95Gas)
                                      or  gas.__class__.__name__ == 'IAPWS95Gas')
                                and (not solids))

        self.V_only_lemmon2000 = (len(liquids) == 0
                                 and (isinstance(gas, DryAirLemmon)
                                      or  gas.__class__.__name__ == 'DryAirLemmon')
                                and (not solids))

        # TODO implement as function of phases/or EOS
        self.VL_only_CEOSs_same = (self.VL_only_CEOSs and
                                   self.liquids[0].eos_class is self.gas.eos_class
                                   # self.liquids[0].kijs == self.gas.kijs
                                   and not (isinstance(self.liquids[0], (IGMIX,))
                                            or isinstance(self.gas, (IGMIX,))) 
                                   and self.VL_EOS_hacks)

        self.VL_only_CoolProp = (len(liquids) == 1 
                                 and isinstance(liquids[0], CoolPropLiquid)
                                 and isinstance(gas, CoolPropGas)
                                 and (not solids) and liquids[0].backend == gas.backend and
                                 liquids[0].fluid == gas.fluid)

        self.VL_IG_activity = (len(liquids) == 1
                               and isinstance(liquids[0], GibbsExcessLiquid)
                               and (isinstance(gas, IdealGas) or gas.eos_class is IGMIX)
                               and len(solids) == 0)

        if self.VL_only_CEOSs_same:
            # Two phase pure eoss are two phase up to the critical point only! Then one phase
            self.eos_pure_STP = gas.eos_mix.to_TPV_pure(T=298.15, P=101325.0, V=None, i=0)


        liquids_to_unique_liquids = []
        unique_liquids, unique_liquid_hashes = [], []
        for i, l in enumerate(liquids):
            h = l.model_hash()
            if h not in unique_liquid_hashes:
                unique_liquid_hashes.append(h)
                unique_liquids.append(l)
                liquids_to_unique_liquids.append(i)
            else:
                liquids_to_unique_liquids.append(unique_liquid_hashes.index(h))
        if gas is not None:
            gas_hash = gas.model_hash(True)

        gas_to_unique_liquid = None
        for i, l in enumerate(liquids):
            h = l.model_hash(True)
            if gas is not None and gas_hash == h:
                gas_to_unique_liquid = liquids_to_unique_liquids[i]
                break

        self.gas_to_unique_liquid = gas_to_unique_liquid
        self.liquids_to_unique_liquids = liquids_to_unique_liquids

        self.unique_liquids = unique_liquids
        self.unique_liquid_count = len(unique_liquids)
        self.unique_phases = [gas] + unique_liquids if gas is not None else unique_liquids
        if solids:
            self.unique_phases += solids
        self.unique_phase_count = (1 if gas is not None else 0) + self.unique_liquid_count + len(solids)
        self.unique_liquid_hashes = unique_liquid_hashes
        self.T_MIN_FLASH = max(p.T_MIN_FLASH for p in self.phases)



    def flash_TPV(self, T, P, V, zs=None, solution=None, hot_start=None):
        betas = [1.0]

        if solution is None:
            fun = lambda obj: obj.G()
        elif solution == 'high':
            fun = lambda obj: -obj.T
        elif solution == 'low':
            fun = lambda obj: obj.T
        elif callable(solution):
            fun = solution
        else:
            raise ValueError("Did not recognize solution %s" %(solution))

        if self.phase_count == 1:
            phase = self.phases[0].to(zs=zs, T=T, P=P, V=V)
            return None, [phase], [], betas, None
        elif self.VL_only_CoolProp:
            sln = self.gas.to(zs, T=T, P=P, V=V, prefer_phase=8)
#            if sln.phase == 'l':
#                return None, [sln], [], betas, None
            return None, [], [sln], betas, None
        elif self.VL_only_CEOSs_same and V is None and solution is None:
            gas = self.gas.to(zs=zs, T=T, P=P, V=V)
            if gas.eos_mix.phase == 'l/g':
                gas.eos_mix.solve_missing_volumes()
                if gas.eos_mix.G_dep_l < gas.eos_mix.G_dep_g:
                    l = self.liquid.to_TP_zs(T, P, zs, other_eos=gas.eos_mix)
                    return None, [l], [], betas, None
                return gas, [], [], betas, None
            elif gas.eos_mix.phase == 'g':
                return gas, [], [], betas, None
            else:
                return None, [gas], [], betas, None
        elif self.VL_IG_activity and self.VL_IG_hack and V is None and solution is None:
            l = self.liquid.to(zs=zs, T=T, P=P, V=V)
            if P > l.Psats()[0]:
                return None, [l], [], betas, None
            else:
                gas = self.gas.to(zs=zs, T=T, P=P, V=V)
                return gas, [], [], betas, None
        elif self.VL_only_CEOSs_same and V is not None and (T is not None or P is not None) and solution is None:
            gas = self.gas.to(zs=zs, T=T, P=P, V=V)
            if gas.eos_mix.phase == 'g':
                return gas, [], [], betas, None
            else:
                return None, [gas], [], betas, None
        elif self.VL_only_IAPWS95 and solution is None:
            if T is not None:
                if T > iapws95_Tc:
                    # super critical no matter what
                    gas = self.gas.to(zs=zs, T=T, P=P, V=V)
                    return gas, [], [], betas, None
                elif P is not None:
                    Psat = iapws95_Psat(T)
                    if P < Psat:
                        gas = self.gas.to(zs=zs, T=T, P=P, V=V)
                        return gas, [], [], betas, None
                    else:
                        l = self.liquid.to(zs=zs, T=T, P=P, V=V)
                        return None, [l], [], betas, None
                elif V is not None:
                    rhol_sat = iapws95_rhol_sat(T)
                    rho_mass = Vm_to_rho(V, iapws95_MW)
                    if rho_mass >= rhol_sat:
                        l = self.liquid.to(zs=zs, T=T, V=V)
                        return None, [l], [], betas, None
                    rhog_sat = iapws95_rhog_sat(T)
                    if rho_mass <= rhog_sat:
                        gas = self.gas.to(zs=zs, T=T, V=V)
                        return gas, [], [], betas, None
                    # There is no feasible solution between the two curves

            elif P is not None and V is not None:
                T = iapws95_T(P=P, rho=Vm_to_rho(V, iapws95_MW))
                try:
                    Tsat = iapws95_Tsat(P)
                    if T < Tsat:
                        l = self.liquid.to(zs=zs, T=T, V=V)
                        return None, [l], [], betas, None
                    else:
                        gas = self.gas.to(zs=zs, T=T, V=V)
                        return gas, [], [], betas, None
                except:
                    l = self.liquid.to(zs=zs, T=T, V=V)
                    return None, [l], [], betas, None
                # TODO more logic

        if self.gas_count:
            gas = self.gas.to(zs=zs, T=T, P=P, V=V)
            G_min, lowest_phase = fun(gas), gas
        else:
            G_min, lowest_phase = 1e100, None
            gas = None

        liquids = []
        for l in self.liquids:
            l = l.to(zs=zs, T=T, P=P, V=V)
            G = fun(l)
            if G < G_min:
                G_min, lowest_phase = G, l
            liquids.append(l)


        solids = []
        for s in self.solids:
            s = s.to(zs=zs, T=T, P=P, V=V)
            G = fun(s)
            if G < G_min:
                G_min, lowest_phase = G, s
            solids.append(s)

        if lowest_phase is gas:
            return lowest_phase, [], [], betas, None
        elif lowest_phase in liquids:
            return None, [lowest_phase], [], betas, None
        else:
            return None, [], [lowest_phase], betas, None

    def Psat_guess(self, T):
        if self.VL_only_CEOSs_same:
            # Two phase pure eoss are two phase up to the critical point only! Then one phase
            Psat = self.eos_pure_STP.Psat(T)
        #
        else:
            try:
                Psat = self.correlations.VaporPressures[0](T)
            except:
                # Last resort
                Psat = 1e5
        return Psat

    def flash_TVF(self, T, VF=None, zs=None, hot_start=None):
        zs = [1.0]
        if self.VL_only_CoolProp:
            if coolprop_phase.caching_state_CoolProp is None:
                coolprop_phase.set_coolprop_constants()
            sat_gas_CoolProp = coolprop_phase.caching_state_CoolProp(self.gas.backend, self.gas.fluid, 1, T, CPQT_INPUTS, CPunknown, None)
            sat_gas = self.gas.from_AS(sat_gas_CoolProp)
            sat_liq = self.liquid.to(zs=zs, T=T, V=1.0/sat_gas_CoolProp.saturated_liquid_keyed_output(CPiDmolar))
            return sat_gas.P, sat_liq, sat_gas, 0, 0.0
        elif self.VL_IG_activity:
            Psat = self.liquid.Psats_at(T)[0]
            sat_gas = self.gas.to_TP_zs(T, Psat, zs)
            sat_liq = self.liquid.to_TP_zs(T, Psat, zs)
            return Psat, sat_liq, sat_gas, 0, 0.0
        elif self.VL_only_IAPWS95:
            if T > iapws95_Tc:
                raise PhaseExistenceImpossible("Specified T is in the supercritical region", zs=zs, T=T)

            Psat = iapws95_Psat(T)
            sat_gas = self.gas.to(T=T, V=rho_to_Vm(iapws95_rhog_sat(T), self.gas._MW), zs=zs)
            sat_liq = self.liquid.to(T=T, V=rho_to_Vm(iapws95_rhol_sat(T), self.liquid._MW), zs=zs)
            return Psat, sat_liq, sat_gas, 0, 0.0
        Psat = self.Psat_guess(T)
        gas = self.gas.to_TP_zs(T, Psat, zs)

        if self.VL_only_CEOSs_same:
            if T > self.constants.Tcs[0]:
                raise PhaseExistenceImpossible("Specified T is in the supercritical region", zs=zs, T=T)

            sat_liq = self.liquids[0].to_TP_zs(T, Psat, zs, other_eos=gas.eos_mix)
            return Psat, sat_liq, gas, 0, 0.0

        liquids = [l.to_TP_zs(T, Psat, zs) for l in self.liquids]
#        return TVF_pure_newton(Psat, T, liquids, gas, maxiter=self.TVF_maxiter, xtol=self.TVF_xtol)
        Psat, l, g, iterations, err = TVF_pure_secant(Psat, T, liquids, gas, maxiter=self.TVF_maxiter, xtol=self.TVF_xtol)
        if l.Z() == g.Z():
            raise PhaseExistenceImpossible("Converged to trivial solution", zs=zs, T=T)

#        print('P', P, 'solved')
        return Psat, l, g, iterations, err

    def flash_PVF(self, P, VF=None, zs=None, hot_start=None):
        zs = [1.0]
        if self.VL_only_CoolProp:
            if coolprop_phase.caching_state_CoolProp is None:
                coolprop_phase.set_coolprop_constants()
            sat_gas_CoolProp = coolprop_phase.caching_state_CoolProp(self.gas.backend, self.gas.fluid, P, 1.0, CPPQ_INPUTS, CPunknown, None)
            sat_gas = self.gas.from_AS(sat_gas_CoolProp)
            sat_liq = self.liquids[0].to(zs=zs, T=sat_gas.T, V=1.0/sat_gas_CoolProp.saturated_liquid_keyed_output(CPiDmolar))
            return sat_gas.T, sat_liq, sat_gas, 0, 0.0
        elif self.VL_only_CEOSs_same:
            if P > self.constants.Pcs[0]:
                raise PhaseExistenceImpossible("Specified P is in the supercritical region", zs=zs, P=P)
            try:
                Tsat = self.eos_pure_STP.Tsat(P)
            except:
                raise PhaseExistenceImpossible("Failed to calculate VL equilibrium T; likely supercritical", zs=zs, P=P)
            sat_gas = self.gas.to_TP_zs(Tsat, P, zs)
            sat_liq = self.liquids[0].to_TP_zs(Tsat, P, zs, other_eos=sat_gas.eos_mix)
            return Tsat, sat_liq, sat_gas, 0, 0.0
        elif self.VL_IG_activity:
            Tsat = self.correlations.VaporPressures[0].solve_property(P)
            sat_gas = self.gas.to_TP_zs(Tsat, P, zs)
            sat_liq = self.liquid.to_TP_zs(Tsat, P, zs)
            return Tsat, sat_liq, sat_gas, 0, 0.0
        elif self.VL_only_IAPWS95:
            if P > iapws95_Pc:
                raise PhaseExistenceImpossible("Specified P is in the supercritical region", zs=zs, P=P)

            Tsat = iapws95_Tsat(P)
            sat_gas = self.gas.to(T=Tsat, V=1e-3*iapws95_MW/iapws95_rhog_sat(Tsat), zs=zs)
            sat_liq = self.liquid.to(T=Tsat, V=1e-3*iapws95_MW/iapws95_rhol_sat(Tsat), zs=zs)
            return Tsat, sat_liq, sat_gas, 0, 0.0
        else:
            Tsat = self.correlations.VaporPressures[0].solve_property(P)
        gas = self.gas.to_TP_zs(Tsat, P, zs)
        liquids = [l.to_TP_zs(Tsat, P, zs) for l in self.liquids]
        Tsat, l, g, iterations, err = PVF_pure_newton(Tsat, P, liquids, gas, maxiter=self.PVF_maxiter, xtol=self.PVF_xtol)
        if l.Z() == g.Z():
            raise PhaseExistenceImpossible("Converged to trivial solution", zs=zs, P=P)
        return Tsat, l, g, iterations, err
#        return PVF_pure_secant(Tsat, P, liquids, gas, maxiter=200, xtol=1E-10)

    def flash_TSF(self, T, SF=None, zs=None, hot_start=None):
        # if under triple point search for gas - otherwise search for liquid
        # For water only there is technically two solutions at some point for both
        # liquid and gas, flag?

        # The solid-liquid interface is NOT working well...
        # Worth getting IAPWS going to compare. Maybe also other EOSs
        if T < self.constants.Tts[0]:
            Psub = self.correlations.SublimationPressures[0](T)
            try_phases = [self.gas] + self.liquids
        else:
            try_phases = self.liquids
            Psub = 1e6

        return TSF_pure_newton(Psub, T, try_phases, self.solids,
                               maxiter=self.TSF_maxiter, xtol=self.TSF_xtol)

    def flash_PSF(self, P, SF=None, zs=None, hot_start=None):
        if P < self.constants.Pts[0]:
            Tsub = self.correlations.SublimationPressures[0].solve_property(P)
            try_phases = [self.gas] + self.liquids
        else:
            try_phases = self.liquids
            Tsub = 1e6

        return PSF_pure_newton(Tsub, P, try_phases, self.solids,
                               maxiter=self.PSF_maxiter, xtol=self.PSF_xtol)


    def flash_double(self, spec_0_val, spec_1_val, spec_0_var, spec_1_var):
        pass


    def flash_TPV_HSGUA_VL_bound_first(self, fixed_var_val, spec_val, fixed_var='P',
                                 spec='H', iter_var='T', hot_start=None,
                                 selection_fun_1P=None, cubic=True):
        constants, correlations = self.constants, self.correlations
        zs = [1.0]
        VL_liq, VL_gas = None, None
        flash_convergence = {}
        has_VL = False
        need_both = True
        if fixed_var == 'T':
            if self.Psat_guess(fixed_var_val) > 1e-2:
                Psat, VL_liq, VL_gas, VL_iter, VL_err = self.flash_TVF(fixed_var_val, VF=.5, zs=zs)
                has_VL = True
        elif fixed_var == 'P':
            if fixed_var_val > 1e-2:
                Tsat, VL_liq, VL_gas, VL_iter, VL_err = self.flash_PVF(fixed_var_val, VF=.5, zs=zs)
                has_VL = True
        if has_VL:
            need_both = False
            spec_val_l = getattr(VL_liq, spec)()
            spec_val_g = getattr(VL_gas, spec)()
            VF = (spec_val - spec_val_l) / (spec_val_g - spec_val_l)
            if 0.0 <= VF <= 1.0:
                return VL_gas, [VL_liq], [], [VF, 1.0 - VF], flash_convergence
            elif VF < 0.0:
                phases = [self.liquid, self.gas]
            else:
                phases = [self.gas, self.liquid]
        else:
            phases = self.phases
        solutions_1P = []
        results_G_min_1P = None
        if hot_start is None:
            last_conv = None
        elif iter_var == 'T':
            last_conv = hot_start.T
        elif iter_var == 'P':
            last_conv = hot_start.P
        for phase in phases:
            try:
                # TODO: use has_VL to bound the solver
                T, P, phase, iterations, err = solve_PTV_HSGUA_1P(phase, zs, fixed_var_val, spec_val, fixed_var=fixed_var,
                                                                  spec=spec, iter_var=iter_var, constants=constants, correlations=correlations, last_conv=last_conv,
                                                                  oscillation_detection=cubic,
                                                                  guess_maxiter=self.TPV_HSGUA_guess_maxiter, guess_xtol=self.TPV_HSGUA_guess_xtol,
                                                                  maxiter=self.TPV_HSGUA_maxiter, xtol=self.TPV_HSGUA_xtol)
                if cubic:
                    phase.eos_mix.solve_missing_volumes()
                    if phase.eos_mix.phase == 'l/g':
                        # Check we are not metastable
                        if min(phase.eos_mix.G_dep_l, phase.eos_mix.G_dep_g) == phase.G_dep(): # If we do not have a metastable phase
                            if isinstance(phase, CEOSGas):
                                g, ls = phase, []
                            else:
                                g, ls = None, [phase]
                            flash_convergence['err'] = err
                            flash_convergence['iterations'] = iterations
                            return g, ls, [], [1.0], flash_convergence
                    else:
                        if isinstance(phase, (CEOSGas, IdealGas)):
                            g, ls = phase, []
                        else:
                            g, ls = None, [phase]
                        flash_convergence['err'] = err
                        flash_convergence['iterations'] = iterations
                        return g, ls, [], [1.0], flash_convergence
                else:
                    if isinstance(phase, (CEOSGas, IdealGas)):
                        g, ls = phase, []
                    else:
                        g, ls = None, [phase]
                    flash_convergence['err'] = err
                    flash_convergence['iterations'] = iterations
                    return g, ls, [], [1.0], flash_convergence

            except Exception as e:
                print(e)
                solutions_1P.append(None)



    def flash_TPV_HSGUA(self, fixed_var_val, spec_val, fixed_var='P', spec='H',
                        iter_var='T', zs=None, solution=None,
                        selection_fun_1P=None, hot_start=None,
                        iter_var_backup=None):
        # Be prepared to have a flag here to handle zero flow
        zs = [1.0]
        constants, correlations = self.constants, self.correlations
        if solution is None:
            if fixed_var == 'P' and spec == 'H':
                fun = lambda obj: -obj.S()
            elif fixed_var == 'P' and spec == 'S':
               # fun = lambda obj: obj.G()
                fun = lambda obj: obj.H() # Michaelson
            elif fixed_var == 'V' and spec == 'U':
                fun = lambda obj: -obj.S()
            elif fixed_var == 'V' and spec == 'S':
                fun = lambda obj: obj.U()
            elif fixed_var == 'P' and spec == 'U':
                fun = lambda obj: -obj.S() # promising
                # fun = lambda obj: -obj.H() # not bad not as good as A
                # fun = lambda obj: obj.A() # Pretty good
                # fun = lambda obj: -obj.V() # First
            else:
                fun = lambda obj: obj.G()
        else:
            if solution == 'high':
                fun = lambda obj: -obj.value(iter_var)
            elif solution == 'low':
                fun = lambda obj: obj.value(iter_var)
            elif callable(solution):
                fun = solution
            else:
                raise ValueError("Unrecognized solution")

        selection_fun_1P_specified = True
        if selection_fun_1P is None:
            selection_fun_1P_specified = False
            def selection_fun_1P(new, prev):
                if fixed_var == 'P' and spec == 'S':
                    if new[-1] < prev[-1]:
                        if new[0] < 1.0 and prev[0] > 1.0:
                            # Found a very low temperature solution do not take it
                            return False
                        return True
                    elif (prev[0] < 1.0 and new[0] > 1.0):
                        return True

                else:
                    if new[-1] < prev[-1]:
                        return True
                return False

        if (self.VL_only_CEOSs_same or self.VL_IG_activity) and not selection_fun_1P_specified and solution is None and fixed_var != 'V':
            try:
                return self.flash_TPV_HSGUA_VL_bound_first(fixed_var_val=fixed_var_val, spec_val=spec_val, fixed_var=fixed_var,
                                     spec=spec, iter_var=iter_var, hot_start=hot_start, selection_fun_1P=selection_fun_1P, cubic=self.VL_only_CEOSs_same)
            except PhaseExistenceImpossible:
                pass
        elif self.V_only_lemmon2000 and not selection_fun_1P_specified and fixed_var == 'V' and iter_var == 'P':
            # Specifically allow the solution to be specified, equation goes wonky around 50000
            iter_var = 'T'
#            if sln is not None:
#                return sln
        try:
            solutions_1P = []
            G_min = 1e100
            results_G_min_1P = None
            for phase in self.phases:
                # TODO: for eoss wit boundaries, and well behaved fluids, only solve ocne instead of twice (i.e. per phase, doubling the computation.)
                try:
                    T, P, phase, iterations, err = solve_PTV_HSGUA_1P(phase, zs, fixed_var_val, spec_val, fixed_var=fixed_var,
                                                                      spec=spec, iter_var=iter_var, constants=constants, correlations=correlations,
                                                                      guess_maxiter=self.TPV_HSGUA_guess_maxiter, guess_xtol=self.TPV_HSGUA_guess_xtol,
                                                                      maxiter=self.TPV_HSGUA_maxiter, xtol=self.TPV_HSGUA_xtol)

                    G = fun(phase)
                    new = [T, phase, iterations, err, G]
                    if results_G_min_1P is None or selection_fun_1P(new, results_G_min_1P):
#                    if G < G_min:
                        G_min = G
                        results_G_min_1P = new

                    solutions_1P.append(new)
                except Exception as e:
#                    print(e)
                    solutions_1P.append(None)
        except:
            pass


        try:
            VL_liq, VL_gas = None, None
            G_VL = 1e100
            # BUG - P IS NOW KNOWN!
            if self.gas_count and self.liquid_count:
                if fixed_var == 'T' and self.Psat_guess(fixed_var_val) > 1e-2:
                    Psat, VL_liq, VL_gas, VL_iter, VL_err = self.flash_TVF(fixed_var_val, zs=zs, VF=.5)
                elif fixed_var == 'P' and fixed_var_val > 1e-2:
                    Tsat, VL_liq, VL_gas, VL_iter, VL_err = self.flash_PVF(fixed_var_val, zs=zs, VF=.5)
            elif fixed_var == 'V':
                raise NotImplementedError("Does not make sense here because there is no actual vapor frac spec")

#                VL_flash = self.flash(P=P, VF=.4)
#            print('hade it', VL_liq, VL_gas)
            spec_val_l = getattr(VL_liq, spec)()
            spec_val_g = getattr(VL_gas, spec)()
#                spec_val_l = getattr(VL_flash.liquid0, spec)()
#                spec_val_g = getattr(VL_flash.gas, spec)()
            VF = (spec_val - spec_val_l)/(spec_val_g - spec_val_l)
            if 0.0 <= VF <= 1.0:
                G_l = fun(VL_liq)
                G_g = fun(VL_gas)
                G_VL = G_g*VF + G_l*(1.0 - VF)
            else:
                VF = None
        except Exception as e:
#            print(e, spec)
            VF = None

        try:
            G_SF = 1e100
            if self.solid_count and (self.gas_count or self.liquid_count):
                VS_flash = self.flash(SF=.5, **{fixed_var: fixed_var_val})
#                VS_flash = self.flash(P=P, SF=1)
                spec_val_s = getattr(VS_flash.solid0, spec)()
                spec_other = getattr(VS_flash.phases[0], spec)()
                SF = (spec_val - spec_val_s)/(spec_other - spec_val_s)
                if SF < 0.0 or SF > 1.0:
                    raise ValueError("Not apply")
                else:
                    G_other = fun(VS_flash.phases[0])
                    G_s = fun(VS_flash.solid0)
                    G_SF = G_s*SF + G_other*(1.0 - SF)
            else:
                SF = None
        except:
            SF = None

        gas_phase = None
        ls = []
        ss = []
        betas = []

        # If a 1-phase solution arrose, set it
        if results_G_min_1P is not None:
            betas = [1.0]
            T, phase, iterations, err, _ = results_G_min_1P
            if phase.is_gas:
                gas_phase = results_G_min_1P[1]
            elif phase.is_liquid:
                ls = [results_G_min_1P[1]]
            elif phase.is_solid:
                ss = [results_G_min_1P[1]]

        flash_convergence = {}
        if G_VL < G_min:
            skip_VL = False

#            if fixed_var == 'P' and spec == 'S' and fixed_var_val < 1.0 and 0:
#                skip_VL = True

            if not skip_VL:
                G_min = G_VL
                ls = [VL_liq]
                gas_phase = VL_gas
                betas = [VF, 1.0 - VF]
                ss = [] # Ensure solid unset
                T = VL_liq.T
                iterations = 0
                err = 0.0
                flash_convergence['VF flash convergence'] = {'iterations': VL_iter, 'err': VL_err}

        if G_SF < G_min:
            try:
                ls = [SF_flash.liquid0]
                gas_phase = None
            except:
                ls = []
                gas_phase = SF_flash.gas
            ss = [SF_flash.solid0]
            betas = [1.0 - SF, SF]
            T = SF_flash.T
            iterations = 0
            err = 0.0
            flash_convergence['SF flash convergence'] = SF_flash.flash_convergence

        if G_min == 1e100:
            '''Calculate the values of val at minimum and maximum temperature
            for each phase.
            Calculate the val at the phase changes.
            Include all in the exception to prove within bounds;
            also have a self check to say whether or not the value should have
            had a converged value.
            '''
            if iter_var == 'T':
                min_bound = Phase.T_MIN_FIXED*(1.0-1e-15)
                max_bound = Phase.T_MAX_FIXED*(1.0+1e-15)
            elif iter_var == 'P':
                min_bound = Phase.P_MIN_FIXED*(1.0-1e-15)
                max_bound = Phase.P_MAX_FIXED*(1.0+1e-15)
            elif iter_var == 'V':
                min_bound = Phase.V_MIN_FIXED*(1.0-1e-15)
                max_bound = Phase.V_MAX_FIXED*(1.0+1e-15)

            phases_at_min = []
            phases_at_max = []

#            specs_at_min = []
#            specs_at_max = []

            had_solution = False
            uncertain_solution = False

            s = ''
            phase_kwargs = {fixed_var: fixed_var_val, 'zs': zs}
            for phase in self.phases:

                try:
                    phase_kwargs[iter_var] = min_bound
                    p = phase.to(**phase_kwargs)
                    phases_at_min.append(p)

                    phase_kwargs[iter_var] = max_bound
                    p = phase.to(**phase_kwargs)
                    phases_at_max.append(p)

                    low, high = getattr(phases_at_min[-1], spec)(), getattr(phases_at_max[-1], spec)()
                    low, high = min(low, high), max(low, high)
                    s += '%s 1 Phase solution: (%g, %g); ' %(p.__class__.__name__, low, high)
                    if low <= spec_val <= high:
                        had_solution = True
                except:
                    uncertain_solution = True

            if VL_liq is not None:
                s += '(%s, %s) VL 2 Phase solution: (%g, %g); ' %(
                        VL_liq.__class__.__name__, VL_gas.__class__.__name__,
                        spec_val_l, spec_val_g)
                VL_min_spec, VL_max_spec = min(spec_val_l, spec_val_g), max(spec_val_l, spec_val_g),
                if VL_min_spec <= spec_val <= VL_max_spec:
                    had_solution = True
            if SF is not None:
                s += '(%s, %s) VL 2 Phase solution: (%g, %g); ' %(
                        VS_flash.phases[0].__class__.__name__, VS_flash.solid0.__class__.__name__,
                        spec_val_s, spec_other)
                S_min_spec, S_max_spec = min(spec_val_s, spec_other), max(spec_val_s, spec_other),
                if S_min_spec <= spec_val <= S_max_spec:
                    had_solution = True
            if had_solution:
                raise UnconvergedError("Could not converge but solution detected in bounds: %s" %s)
            elif uncertain_solution:
                raise UnconvergedError("Could not converge and unable to detect if solution detected in bounds")
            else:
                raise NoSolutionError("No physical solution in bounds for %s=%s at %s=%s: %s" %(spec, spec_val, fixed_var, fixed_var_val, s))

        flash_convergence['iterations'] = iterations
        flash_convergence['err'] = err

        return gas_phase, ls, ss, betas, flash_convergence

    def compare_flashes(self, state, inputs=None):
        # do a PT
        PT = self.flash(T=state.T, P=state.P)

        if inputs is None:
            inputs = [('T', 'P'),
                                 ('T', 'V'),
                                 ('P', 'V'),

                                 ('T', 'H'),
                                 ('T', 'S'),
                                 ('T', 'U'),

                                 ('P', 'H'),
                                 ('P', 'S'),
                                 ('P', 'U'),

                                 ('V', 'H'),
                                 ('V', 'S'),
                                 ('V', 'U')]

        states = []
        for p0, p1 in inputs:
            kwargs = {}

            p0_spec = getattr(state, p0)
            try:
                p0_spec = p0_spec()
            except:
                pass
            p1_spec = getattr(state, p1)
            try:
                p1_spec = p1_spec()
            except:
                pass
            kwargs = {}
            kwargs[p0] = p0_spec
            kwargs[p1] = p1_spec
            new = self.flash(**kwargs)
            states.append(new)
        return states

    def assert_flashes_same(self, reference, states, props=['T', 'P', 'V', 'S', 'H', 'G', 'U', 'A'], rtol=1e-7):
        ref_props = [reference.value(k) for k in props]
        for i, k in enumerate(props):
            ref = ref_props[i]
            for s in states:
                assert_close(s.value(k), ref, rtol=rtol)

    def generate_VF_data(self, Pmin=None, Pmax=None, pts=100,
                         props=['T', 'P', 'V', 'S', 'H', 'G', 'U', 'A']):
        '''Could use some better algorithms for generating better data? Some of
        the solutions count on this.
        '''
        Pc = self.constants.Pcs[0]
        if Pmax is None:
            Pmax = Pc
        if Pmin is None:
            Pmin = 1e-2
        if self.VL_only_CoolProp:
            AS = self.gas.AS
            Pmin = AS.trivial_keyed_output(CPiP_min)*(1.0 + 1e-3)
            Pmax = AS.p_critical()*(1.0 - 1e-7)

        Tmin, liquid, gas, iters, flash_err = self.flash_PVF(P=Pmin, VF=.5, zs=[1.0])
        Tmax, liquid, gas, iters, flash_err = self.flash_PVF(P=Pmax, VF=.5, zs=[1.0])

        liq_props, gas_props = [[] for _ in range(len(props))], [[] for _ in range(len(props))]
        # Lots of issues near Tc - split the range into low T and high T
        T_mid = 0.1*Tmin + 0.95*Tmax
        T_next = 0.045*Tmin + 0.955*Tmax

        Ts = linspace(Tmin, T_mid, pts//2)
        Ts += linspace(T_next, Tmax, pts//2)
        Ts.insert(-1, Tmax*(1-1e-8))
        Ts.sort()
        for T in Ts:
            Psat, liquid, gas, iters, flash_err = self.flash_TVF(T, VF=.5, zs=[1.0])
            for i, prop in enumerate(props):
                liq_props[i].append(liquid.value(prop))
                gas_props[i].append(gas.value(prop))

        return liq_props, gas_props

    def build_VF_interpolators(self, T_base=True, P_base=True, pts=50):
        self.liq_VF_interpolators = liq_VF_interpolators = {}
        self.gas_VF_interpolators = gas_VF_interpolators = {}
        props = ['T', 'P', 'V', 'S', 'H', 'G', 'U', 'A',
                 'dS_dT', 'dH_dT', 'dG_dT', 'dU_dT', 'dA_dT',
                 'dS_dP', 'dH_dP', 'dG_dP', 'dU_dP', 'dA_dP',
                 'fugacity', 'dfugacity_dT', 'dfugacity_dP']

        liq_props, gas_props = self.generate_VF_data(props=props, pts=pts)
        self.liq_VF_data = liq_props
        self.gas_VF_data = gas_props
        self.props_VF_data = props

        if T_base and P_base:
            base_props, base_idxs = ('T', 'P'), (0, 1)
        elif T_base:
            base_props, base_idxs = ('T',), (0,)
        elif P_base:
            base_props, base_idxs = ('P',), (1,)

        self.VF_data_base_props = base_props
        self.VF_data_base_idxs = base_idxs

        self.VF_data_spline_kwargs = spline_kwargs = dict(bc_type='natural', extrapolate=False)

        try:
            self.build_VF_splines()
        except:
            pass

    def build_VF_splines(self):
        self.VF_interpolators_built = True
        props = self.props_VF_data
        liq_props, gas_props = self.liq_VF_data, self.gas_VF_data
        VF_data_spline_kwargs = self.VF_data_spline_kwargs
        liq_VF_interpolators = self.liq_VF_interpolators
        gas_VF_interpolators = self.gas_VF_interpolators
        from scipy.interpolate import CubicSpline

        for base_prop, base_idx in zip(self.VF_data_base_props, self.VF_data_base_idxs):
            xs = liq_props[base_idx]
            for i, k in enumerate(props):
                if i == base_idx:
                    continue
                try:
                    spline = CubicSpline(xs, liq_props[i], **VF_data_spline_kwargs)
                    liq_VF_interpolators[(base_prop, k)] = spline
                except:
                    pass

                try:
                    spline = CubicSpline(xs, gas_props[i], **VF_data_spline_kwargs)
                    gas_VF_interpolators[(base_prop, k)] = spline
                except:
                    pass



    def flash_VF_HSGUA(self, fixed_var_val, spec_val, fixed_var='VF', spec_var='H', zs=None,
                       hot_start=None, solution='high'):
        # solution at high T by default
        if not self.VF_interpolators_built:
            self.build_VF_interpolators()
        iter_var = 'T' # hardcoded -
        # to make code generic try not to use eos stuff
#        liq_obj = self.liq_VF_interpolators[(iter_var, spec_var)]
#        gas_obj = self.liq_VF_interpolators[(iter_var, spec_var)]
        # iter_var must always be T
        VF = fixed_var_val
        props = self.props_VF_data
        liq_props = self.liq_VF_data
        gas_props = self.gas_VF_data
        iter_idx = props.index(iter_var)
        spec_idx = props.index(spec_var)

        T_idx, P_idx = props.index('T'), props.index('P')
        Ts, Ps = liq_props[T_idx], liq_props[P_idx]

        dfug_dT_idx = props.index('dfugacity_dT')
        dfug_dP_idx = props.index('dfugacity_dP')

        dspec_dT_var = 'd%s_dT' %(spec_var)
        dspec_dP_var = 'd%s_dP' %(spec_var)
        dspec_dT_idx = props.index(dspec_dT_var)
        dspec_dP_idx = props.index(dspec_dP_var)

        bounding_idx, bounding_Ts = [], []

        spec_values = []
        dspec_values = []

        d_sign_changes = False
        d_sign_changes_idx = []

        for i in range(len(liq_props[0])):
            v = liq_props[spec_idx][i]*(1.0 - VF) + gas_props[spec_idx][i]*VF

            dfg_T, dfl_T = gas_props[dfug_dT_idx][i], liq_props[dfug_dT_idx][i]
            dfg_P, dfl_P = gas_props[dfug_dP_idx][i], liq_props[dfug_dP_idx][i]
            at_critical = False
            try:
                dPsat_dT = (dfg_T - dfl_T)/(dfl_P - dfg_P)
            except ZeroDivisionError:
                at_critical = True
                dPsat_dT = self.constants.Pcs[0] #

            dv_g = dPsat_dT*gas_props[dspec_dP_idx][i] + gas_props[dspec_dT_idx][i]
            dv_l = dPsat_dT*liq_props[dspec_dP_idx][i] + liq_props[dspec_dT_idx][i]
            dv = dv_l*(1.0 - VF) + dv_g*VF
            if at_critical:
                dv = dspec_values[-1]

            if i > 0:
                if ((v <= spec_val <= spec_values[-1]) or (spec_values[-1] <= spec_val <= v)):
                    bounding_idx.append((i-1, i))
                    bounding_Ts.append((Ts[i-1], Ts[i]))

                if dv*dspec_values[-1] < 0.0:
                    d_sign_changes = True
                    d_sign_changes_idx.append((i-1, i))

            spec_values.append(v)
            dspec_values.append(dv)

        # if len(bounding_idx) < 2 and d_sign_changes:
        # Might not be in the range where there are multiple solutions
        #     raise ValueError("Derivative sign changes but only found one bounding value")


        # if len(bounding_idx) == 1:
        if len(bounding_idx) == 1 and (not d_sign_changes or (bounding_idx != d_sign_changes_idx and 1)):
            # Not sure about condition
            # Go right for the root
            T_low, T_high = bounding_Ts[0][0], bounding_Ts[0][1]
            idx_low, idx_high = bounding_idx[0][0], bounding_idx[0][1]

            val_low, val_high = spec_values[idx_low], spec_values[idx_high]
            dval_low, dval_high = dspec_values[idx_low], dspec_values[idx_high]
        # elif len(bounding_idx) == 0 and d_sign_changes:
            # root must be in interval derivative changes: Go right for the root
            # idx_low, idx_high = d_sign_changes_idx[0][0], d_sign_changes_idx[0][1]
            # T_low, T_high = Ts[idx_low], Ts[idx_high]
            #
            # val_low, val_high = spec_values[idx_low], spec_values[idx_high]
            # dval_low, dval_high = dspec_values[idx_low], dspec_values[idx_high]
        elif len(bounding_idx) == 2:
            # pick range and go for it
            if solution == 'high' or solution is None:
                T_low, T_high = bounding_Ts[1][0], bounding_Ts[1][1]
                idx_low, idx_high = bounding_idx[1][0], bounding_idx[1][1]
            else:
                T_low, T_high = bounding_Ts[0][0], bounding_Ts[0][1]
                idx_low, idx_high = bounding_idx[0][0], bounding_idx[0][1]

            val_low, val_high = spec_values[idx_low], spec_values[idx_high]
            dval_low, dval_high = dspec_values[idx_low], dspec_values[idx_high]

        elif (len(bounding_idx) == 1 and d_sign_changes) or (len(bounding_idx) == 0 and d_sign_changes):
            # Gotta find where derivative root changes, then decide if we have two solutions or just one; decide which to pursue
            idx_low, idx_high = d_sign_changes_idx[0][0], d_sign_changes_idx[0][1]
            T_low, T_high = Ts[idx_low], Ts[idx_high]

            T_guess = 0.5*(T_low +T_high)
            T_der_zero, v_zero = self._VF_HSGUA_der_root(T_guess, T_low, T_high, fixed_var_val, spec_val, fixed_var=fixed_var,
                                        spec_var=spec_var)
            high, low = False, False
            if (v_zero < spec_val < spec_values[idx_high]) or (spec_values[idx_high] < spec_val < v_zero):
                high = True
            if (spec_values[idx_low] < spec_val < v_zero) or (v_zero < spec_val < spec_values[idx_low]):
                low = True
            if not low and not high:
                # There was no other solution where the derivative changed
                T_low, T_high = bounding_Ts[0][0], bounding_Ts[0][1]
                idx_low, idx_high = bounding_idx[0][0], bounding_idx[0][1]

                val_low, val_high = spec_values[idx_low], spec_values[idx_high]
                dval_low, dval_high = dspec_values[idx_low], dspec_values[idx_high]
            elif (high and solution == 'high') or not low:
                val_low, val_high = v_zero, spec_values[idx_high]
                dval_low, dval_high = dspec_values[idx_high], dspec_values[idx_high]
                T_low, T_high = T_der_zero, Ts[idx_high]
            else:
                val_low, val_high = spec_values[idx_low], v_zero
                dval_low, dval_high = dspec_values[idx_low], dspec_values[idx_low]
                T_low, T_high = Ts[idx_low], T_der_zero
        elif len(bounding_idx) >2:
            # Entropy plot has 3 solutions, two derivative changes - give up by that point
            if isinstance(solution, int):
                sln_idx = solution
            else:
                sln_idx = {'high': -1, 'mid': -2, 'low': 0}[solution]
            T_low, T_high = bounding_Ts[sln_idx][0], bounding_Ts[sln_idx][1]
            idx_low, idx_high = bounding_idx[sln_idx][0], bounding_idx[sln_idx][1]

            val_low, val_high = spec_values[idx_low], spec_values[idx_high]
            dval_low, dval_high = dspec_values[idx_low], dspec_values[idx_high]

        else:
            raise ValueError("What")

        T_guess_low  = T_low - (val_low - spec_val)/dval_low
        T_guess_high  = T_high - (val_high - spec_val)/dval_high

        if T_low < T_guess_low < T_high and T_low < T_guess_high < T_high:
            T_guess = 0.5*(T_guess_low + T_guess_high)
        else:
            T_guess = 0.5*(T_low + T_high)

        return self.flash_VF_HSGUA_bounded(T_guess, T_low, T_high, fixed_var_val, spec_val, fixed_var=fixed_var, spec_var=spec_var)

    def _VF_HSGUA_der_root(self, guess, low, high, fixed_var_val, spec_val, fixed_var='VF', spec_var='H'):
        dspec_dT_var = 'd%s_dT' % (spec_var)
        dspec_dP_var = 'd%s_dP' % (spec_var)
        VF = fixed_var_val

        val_cache = [None, 0]

        def to_solve(T):
            Psat, liquid, gas, iters, flash_err = self.flash_TVF(T=T, VF=VF, zs=[1.0])
            # Error
            calc_spec_val = getattr(gas, spec_var)()*VF + getattr(liquid, spec_var)()*(1.0 - VF)
            val_cache[0] = calc_spec_val
            val_cache[1] += 1

            dfg_T, dfl_T = gas.dfugacity_dT(), liquid.dfugacity_dT()
            dfg_P, dfl_P = gas.dfugacity_dP(), liquid.dfugacity_dP()
            dPsat_dT = (dfg_T - dfl_T) / (dfl_P - dfg_P)

            dv_g = dPsat_dT*getattr(gas, dspec_dP_var)() + getattr(gas, dspec_dT_var)()
            dv_l = dPsat_dT*getattr(liquid, dspec_dP_var)() + getattr(liquid, dspec_dT_var)()
            dv = dv_l*(1.0 - VF) + dv_g*VF

            return dv

        # import matplotlib.pyplot as plt
        # xs = linspace(low, high, 1000)
        # ys = [to_solve(x) for x in xs]
        # plt.plot(xs, ys)
        # plt.show()
        try:
            T_zero = secant(to_solve, guess, low=low, high=high, xtol=1e-12, bisection=True)
        except:
            T_zero = brenth(to_solve, low, high, xtol=1e-12)
        return T_zero, val_cache[0]

    def flash_VF_HSGUA_bounded(self, guess, low, high, fixed_var_val, spec_val, fixed_var='VF', spec_var='H'):
        dspec_dT_var = 'd%s_dT' % (spec_var)
        dspec_dP_var = 'd%s_dP' % (spec_var)
        VF = fixed_var_val

        cache = [0]
        fprime = True
        def to_solve(T):
            Psat, liquid, gas, iters, flash_err = self.flash_TVF(T=T, VF=VF, zs=[1.0])
            # Error
            calc_spec_val = getattr(gas, spec_var)()*VF + getattr(liquid, spec_var)()*(1.0 - VF)
            err = calc_spec_val - spec_val
            cache[:] = [T, Psat, liquid, gas, iters, flash_err, err, cache[-1]+1]
            if not fprime:
                return err
            # Derivative
            dfg_T, dfl_T = gas.dfugacity_dT(), liquid.dfugacity_dT()
            dfg_P, dfl_P = gas.dfugacity_dP(), liquid.dfugacity_dP()
            dPsat_dT = (dfg_T - dfl_T) / (dfl_P - dfg_P)

            dv_g = dPsat_dT*getattr(gas, dspec_dP_var)() + getattr(gas, dspec_dT_var)()
            dv_l = dPsat_dT*getattr(liquid, dspec_dP_var)() + getattr(liquid, dspec_dT_var)()
            dv = dv_l*(1.0 - VF) + dv_g*VF

            return err, dv

        #
        try:
            T_calc = newton(to_solve, guess, fprime=True, low=low, high=high, xtol=1e-12, require_eval=True)
        except:
            # Zero division error in derivative mostly
            fprime = False
            T_calc = secant(to_solve, guess, low=low, high=high, xtol=1e-12, ytol=guess*1e-5, require_eval=True)


        return cache





    def debug_TVF(self, T, VF=None, pts=2000):
        zs = [1]
        gas = self.gas
        liquids = self.liquids

        def to_solve_newton(P):
            g = gas.to_TP_zs(T, P, zs)
            fugacity_gas = g.fugacities()[0]
            dfugacities_dP_gas = g.dfugacities_dP()[0]
            ls = [l.to_TP_zs(T, P, zs) for l in liquids]
            G_min, lowest_phase = 1e100, None
            for l in ls:
                G = l.G()
                if G < G_min:
                    G_min, lowest_phase = G, l

            fugacity_liq = lowest_phase.fugacities()[0]
            dfugacities_dP_liq = lowest_phase.dfugacities_dP()[0]

            err = fugacity_liq - fugacity_gas
            derr_dP = dfugacities_dP_liq - dfugacities_dP_gas
            return err, derr_dP

        import matplotlib.pyplot as plt
        import numpy as np

        Psat = self.correlations.VaporPressures[0](T)
        Ps = np.hstack([np.logspace(np.log10(Psat/2), np.log10(Psat*2), int(pts/2)),
                        np.logspace(np.log10(1e-6), np.log10(1e9), int(pts/2))])
        Ps = np.sort(Ps)
        values = np.array([to_solve_newton(P)[0] for P in Ps])
        values[values == 0] = 1e-10 # Make them show up on the plot

        plt.loglog(Ps, values, 'x', label='Positive errors')
        plt.loglog(Ps, -values, 'o', label='Negative errors')
        plt.legend(loc='best', fancybox=True, framealpha=0.5)
        plt.show()

    def debug_PVF(self, P, VF=None, pts=2000):
        zs = [1]
        gas = self.gas
        liquids = self.liquids

        def to_solve_newton(T):
            g = gas.to_TP_zs(T, P, zs)
            fugacity_gas = g.fugacities()[0]
            dfugacities_dT_gas = g.dfugacities_dT()[0]
            ls = [l.to_TP_zs(T, P, zs) for l in liquids]
            G_min, lowest_phase = 1e100, None
            for l in ls:
                G = l.G()
                if G < G_min:
                    G_min, lowest_phase = G, l

            fugacity_liq = lowest_phase.fugacities()[0]
            dfugacities_dT_liq = lowest_phase.dfugacities_dT()[0]

            err = fugacity_liq - fugacity_gas
            derr_dT = dfugacities_dT_liq - dfugacities_dT_gas
            return err, derr_dT

        import matplotlib.pyplot as plt
        Psat_obj = self.correlations.VaporPressures[0]

        Tsat = Psat_obj.solve_property(P)
        Tmax = Psat_obj.Tmax
        Tmin = Psat_obj.Tmin


        Ts = np.hstack([np.linspace(Tmin, Tmax, int(pts/4)),
                        np.linspace(Tsat-30, Tsat+30, int(pts/4))])
        Ts = np.sort(Ts)

        values = np.array([to_solve_newton(T)[0] for T in Ts])

        plt.semilogy(Ts, values, 'x', label='Positive errors')
        plt.semilogy(Ts, -values, 'o', label='Negative errors')


        min_index = np.argmin(np.abs(values))

        T = Ts[min_index]
        Ts2 = np.linspace(T*.999, T*1.001, int(pts/2))
        values2 = np.array([to_solve_newton(T)[0] for T in Ts2])
        plt.semilogy(Ts2, values2, 'x', label='Positive Fine')
        plt.semilogy(Ts2, -values2, 'o', label='Negative Fine')

        plt.legend(loc='best', fancybox=True, framealpha=0.5)
        plt.show()



    # ph - iterate on PT
    # if oscillating, take those two phases, solve, then get VF
    # other strategy - guess phase, solve h, PT at point to vonfirm!
    # For one phase - solve each phase for H, if there is a solution.
    # Take the one with lowest Gibbs energy




# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from fluids.numerics import assert_close, assert_close1d, assert_close2d, linspace
import numpy as np
import pytest
import json
import pandas as pd
from thermo.volume import *
from thermo.volume import VDI_TABULAR
from thermo.eos import *
import chemicals
from chemicals.utils import Vm_to_rho, zs_to_ws
from thermo.vapor_pressure import VaporPressure
from thermo.utils import POLY_FIT, EXP_POLY_FIT
from chemicals.volume import *
from thermo.volume import LALIBERTE, COSTALD_MIXTURE_FIT, RACKETT_PARAMETERS, COSTALD_MIXTURE, LINEAR, RACKETT
from thermo.volume import HTCOSTALD, COOLPROP, DIPPR_PERRY_8E, VDI_TABULAR, RACKETTFIT, YEN_WOODS_SAT, BHIRUD_NORMAL, VDI_PPDS, TOWNSEND_HALES, HTCOSTALDFIT, CAMPBELL_THODOS, MMSNM0, RACKETT, YAMADA_GUNN
from thermo.volume import PITZER_CURL, EOS, TSONOPOULOS_EXTENDED, ABBOTT, COOLPROP, IDEAL, TSONOPOULOS, CRC_VIRIAL
from thermo.volume import EOS, LINEAR, IDEAL
from thermo.eos_mix import PRMIX
from thermo.coolprop import has_CoolProp

@pytest.mark.CoolProp
@pytest.mark.meta_T_dept
@pytest.mark.skipif(not has_CoolProp(), reason='CoolProp is missing')
def test_VolumeGas_CoolProp():
    SO2 = VolumeGas(CASRN='7446-09-5', MW=64.0638,  Tc=430.8, Pc=7884098.25, omega=0.251, dipole=1.63)
    SO2.method_P = COOLPROP
    assert_close(SO2.TP_dependent_property(305, 1E5), 0.024971467450477493, rtol=1e-13)

@pytest.mark.meta_T_dept
def test_VolumeGas():
    eos = [PR(T=300, P=1E5, Tc=430.8, Pc=7884098.25, omega=0.251)]
    SO2 = VolumeGas(CASRN='7446-09-5', MW=64.0638,  Tc=430.8, Pc=7884098.25, omega=0.251, dipole=1.63, eos=eos)

    SO2.method_P = ABBOTT
    assert_close(SO2.TP_dependent_property(305, 1E5), 0.024995877483424613, rtol=1e-13)

    SO2.method_P = PITZER_CURL
    assert_close(SO2.TP_dependent_property(305, 1E5), 0.024996281566986505, rtol=1e-13)

    SO2.method_P = EOS
    assert_close(SO2.TP_dependent_property(305, 1E5), 0.02502431104578071, rtol=1e-13)

    SO2.method_P = TSONOPOULOS_EXTENDED
    assert_close(SO2.TP_dependent_property(305, 1E5), 0.02499979467057479, rtol=1e-13)

    SO2.method_P = IDEAL
    assert_close(SO2.TP_dependent_property(305, 1E5), 0.02535911098536738, rtol=1e-13)


    # Test that methods return None
    for i in SO2.all_methods_P:
        SO2.method_P = i
        assert SO2.TP_dependent_property(-100, 1E5) is None
        assert SO2.TP_dependent_property(100, -1E5) is None

    with pytest.raises(Exception):
        SO2.test_method_validity_P(300, 1E5, 'BADMETHOD')



    # Ethanol data, calculated from CoolProp
    EtOH = VolumeGas(MW=46.06844, Tc=514.0, Pc=6137000.0, omega=0.635, dipole=1.44, CASRN='64-17-5')
    Ts = [400, 500, 600]
    Ps = [5E3, 1E4, 5E4]

    TP_data = [[0.6646136629870959, 0.8312205608372635, 0.9976236498685168], [0.33203434186506636, 0.4154969056659669, 0.49875532241189763], [0.06596765735649, 0.08291758227639608, 0.09966060658661156]]
    EtOH.add_tabular_data_P(Ts, Ps, TP_data, name='CPdata')
    recalc_pts = [[EtOH.TP_dependent_property(T, P) for T in Ts] for P in Ps]
    assert_close2d(TP_data, recalc_pts)

    assert_close(EtOH.TP_dependent_property(300, 9E4), 0.06596765735649)
    EtOH.tabular_extrapolation_permitted = False
    assert None == EtOH.TP_dependent_property(300, 9E4)

    # Test CRC Virial data
    H2 = VolumeGas(CASRN='1333-74-0')
    H2.method_P = CRC_VIRIAL
    assert_close(H2.TP_dependent_property(300, 1E5), 0.024958843346854165)

@pytest.mark.meta_T_dept
def test_VolumeGas_Tabular_json_eval():
    obj = VolumeGas(MW=46.06844, Tc=514.0, Pc=6137000.0, omega=0.635, dipole=1.44, CASRN='64-17-5')
    Ts = [400, 500, 600]
    Ps = [5E3, 1E4, 5E4]

    TP_data = [[0.6646136629870959, 0.8312205608372635, 0.9976236498685168], [0.33203434186506636, 0.4154969056659669, 0.49875532241189763], [0.06596765735649, 0.08291758227639608, 0.09966060658661156]]
    obj.add_tabular_data_P(Ts, Ps, TP_data, name='CPdata')
    hash0 = hash(obj)
    obj2 = eval(str(obj))
    assert obj2 == obj
    assert hash(obj) == hash0
    assert hash0 == hash(obj2)

    obj2 = VolumeGas.from_json(json.loads(json.dumps(obj.as_json())))
    assert obj == obj2
    assert hash(obj) == hash0
    assert hash(obj2) == hash0

@pytest.mark.CoolProp
@pytest.mark.meta_T_dept
@pytest.mark.skipif(not has_CoolProp(), reason='CoolProp is missing')
def test_VolumeLiquid_CoolProp():
    # Ethanol, test all methods at once
    EtOH = VolumeLiquid(MW=46.06844, Tb=351.39, Tc=514.0, Pc=6137000.0, Vc=0.000168, Zc=0.24125, omega=0.635, dipole=1.44, CASRN='64-17-5', Psat=7882.16)
    EtOH.method = COOLPROP
    assert_close(EtOH.T_dependent_property(305.), 5.912157674597306e-05, rtol=1e-7)

    # Ethanol compressed
    assert [False, True] == [EtOH.test_method_validity_P(300, P, 'COOLPROP') for P in (1E3, 1E5)]
    assert [True, True] == [EtOH.test_method_validity_P(300, P, 'COSTALD_COMPRESSED') for P in (1E3, 1E5)]

    EtOH.method = DIPPR_PERRY_8E
    assert_close(EtOH.calculate_P(298.15, 1E6, 'COSTALD_COMPRESSED'), 5.859498172626399e-05)
    assert_close(EtOH.calculate_P(298.15, 1E6, 'COOLPROP'), 5.861494869071554e-05)



@pytest.mark.meta_T_dept
def test_VolumeLiquid():
    # Ethanol, test all methods at once
    EtOH = VolumeLiquid(MW=46.06844, Tb=351.39, Tc=514.0, Pc=6137000.0, Vc=0.000168, Zc=0.24125, omega=0.635, dipole=1.44, CASRN='64-17-5')
    methods = list(EtOH.all_methods)
    methods.remove(VDI_TABULAR)

    EtOH.method = HTCOSTALD
    assert_close(EtOH.T_dependent_property(305.), 5.526836182702171e-05, rtol=1e-7)
    EtOH.method = DIPPR_PERRY_8E
    assert_close(EtOH.T_dependent_property(305.), 5.9082910221835385e-05, rtol=1e-7)
    EtOH.method = VDI_TABULAR
    assert_close(EtOH.T_dependent_property(305.), 5.9096619917816446e-05, rtol=1e-7)
    EtOH.method = RACKETTFIT
    assert_close(EtOH.T_dependent_property(305.), 5.9680813323376087e-05, rtol=1e-7)
    EtOH.method = YEN_WOODS_SAT
    assert_close(EtOH.T_dependent_property(305.), 5.7847606608322876e-05, rtol=1e-7)
    EtOH.method = BHIRUD_NORMAL
    assert_close(EtOH.T_dependent_property(305.), 5.5070775827335304e-05, rtol=1e-7)
    EtOH.method = VDI_PPDS
    assert_close(EtOH.T_dependent_property(305.), 5.9053167412065864e-05, rtol=1e-7)
    EtOH.method = TOWNSEND_HALES
    assert_close(EtOH.T_dependent_property(305.), 5.4848470492414296e-05, rtol=1e-7)
    EtOH.method = HTCOSTALDFIT
    assert_close(EtOH.T_dependent_property(305.), 5.759457172850207e-05, rtol=1e-7)
    EtOH.method = CAMPBELL_THODOS
    assert_close(EtOH.T_dependent_property(305.), 5.192179422611158e-05, rtol=1e-7)
    EtOH.method = MMSNM0
    assert_close(EtOH.T_dependent_property(305.), 5.821947224585489e-05, rtol=1e-7)
    EtOH.method = RACKETT
    assert_close(EtOH.T_dependent_property(305.), 5.594759690537813e-05, rtol=1e-7)
    EtOH.method = YAMADA_GUNN
    assert_close(EtOH.T_dependent_property(305.), 5.3338200313560806e-05, rtol=1e-7)

    # Test that methods return None
    EtOH = VolumeLiquid(MW=46.06844, Tb=351.39, Tc=514.0, Pc=6137000.0, Vc=0.000168, Zc=0.24125, omega=0.635, dipole=1.44, CASRN='64-17-5')
    EtOH.T_dependent_property(305.) # Initialize the sorted_valid_methods
    EtOH.tabular_extrapolation_permitted = False
    EtOH.extrapolation = None

    for i in list(EtOH.all_methods):
        EtOH.method = i
        assert EtOH.T_dependent_property(600) is None

    EtOH.method = VDI_TABULAR
    EtOH.extrapolation = 'interp1d'
    assert_close(EtOH.T_dependent_property(700.), 0.0005648005718236466)

    with pytest.raises(Exception):
        EtOH.test_method_validity(300, 'BADMETHOD')

    SnCl4 = VolumeLiquid(CASRN='7646-78-8')
    Vm_calcs = []
    for i in SnCl4.all_methods:
        SnCl4.method = i
        Vm_calcs.append(SnCl4.T_dependent_property(305.))

    # Get MMC parameter
    SO2 = VolumeLiquid(MW=64.0638, Tb=263.1, Tc=430.8, Pc=7884098.25, Vc=0.000122, Zc=0.26853, omega=0.251, dipole=1.63, CASRN='7446-09-5', extrapolation=None)
    SO2.method = HTCOSTALD
    assert_close(SO2.T_dependent_property(200.), 4.062166881078707e-05, rtol=1e-7)
    SO2.method = DIPPR_PERRY_8E
    assert_close(SO2.T_dependent_property(200.), 3.965944731935354e-05, rtol=1e-7)
    SO2.method = RACKETTFIT
    assert_close(SO2.T_dependent_property(200.), 3.993452831949474e-05, rtol=1e-7)
    SO2.method = YEN_WOODS_SAT
    assert_close(SO2.T_dependent_property(200.), 4.109189955368029e-05, rtol=1e-7)
    SO2.method = BHIRUD_NORMAL
    assert_close(SO2.T_dependent_property(200.), 4.0604474439638224e-05, rtol=1e-7)
    SO2.method = VDI_PPDS
    assert_close(SO2.T_dependent_property(200.), 3.9697664371887463e-05, rtol=1e-7)
    SO2.method = TOWNSEND_HALES
    assert_close(SO2.T_dependent_property(200.), 4.0256041843356724e-05, rtol=1e-7)
    SO2.method = HTCOSTALDFIT
    assert_close(SO2.T_dependent_property(200.), 3.994849780626379e-05, rtol=1e-7)
    SO2.method = CAMPBELL_THODOS
    assert_close(SO2.T_dependent_property(200.), 3.7484830996072615e-05, rtol=1e-7)
    SO2.method = MMSNM0
    assert_close(SO2.T_dependent_property(200.), 4.0948267317531393e-05, rtol=1e-7)
    SO2.method = RACKETT
    assert_close(SO2.T_dependent_property(200.), 4.060688369271625e-05, rtol=1e-7)
    SO2.method = YAMADA_GUNN
    assert_close(SO2.T_dependent_property(200.), 4.0608202974188125e-05, rtol=1e-7)


    # Get CRC Inorganic invalid
    U = VolumeLiquid(CASRN='7440-61-1')
    assert_close(U.T_dependent_property(1420), 1.3758901734104049e-05)
    assert False == U.test_method_validity(1430, U.method)

    # Test lower limit for BHIRUD_NORMAL
    fake = VolumeLiquid(Tc=1000)
    assert False == fake.test_method_validity(349.5, 'BHIRUD_NORMAL')



    # Ethanol data, calculated from CoolProp
    Ts = [275, 300, 350]
    Ps = [1E5, 5E5, 1E6]
    TP_data = [[5.723868454722602e-05, 5.879532690584185e-05, 6.242643879647073e-05], [5.721587962307716e-05, 5.8767676310784456e-05, 6.238352991918547e-05], [5.718753361659462e-05, 5.8733341196316644e-05, 6.23304095312013e-05]]
    EtOH.add_tabular_data_P(Ts, Ps, TP_data, name='CPdata')
    recalc_pts = [[EtOH.TP_dependent_property(T, P) for T in Ts] for P in Ps]
    assert_close2d(TP_data, recalc_pts)

    EtOH.tabular_extrapolation_permitted = True
    assert_close(EtOH.TP_dependent_property(274, 9E4), 5.723868454722602e-05)
    EtOH.tabular_extrapolation_permitted = False
    assert None == EtOH.TP_dependent_property(300, 9E4)


    with pytest.raises(Exception):
        EtOH.test_method_validity_P(300, 1E5, 'BADMETHOD')


@pytest.mark.meta_T_dept
def test_VolumeLiquid_eval_not_duplicate_VDI_tabular():
    assert 'tabular_data' not in str(VolumeLiquid(CASRN="109-66-0"))

@pytest.mark.meta_T_dept
def test_VolumeLiquidPolynomialTmin():
    # toluene
    v = VolumeLiquid(poly_fit=(178.01, 581.75, [2.2801490297347937e-23, -6.411956871696508e-20, 7.723152902379232e-17, -5.197203733189603e-14, 2.1348482785660093e-11, -5.476649499770259e-09, 8.564670053875876e-07, -7.455178589434267e-05, 0.0028545812080104068]))
    assert v._Tmin_T_trans == v.poly_fit_Tmin
    assert_close1d(v.poly_fit_Tmin_quadratic, [2.159638355215081e-10, 0.0, 8.76710398245817e-05])

    # methanol
    v = VolumeLiquid(poly_fit=(175.7, 502.5, [3.5725079384600736e-23, -9.031033742820083e-20, 9.819637959370411e-17, -5.993173551565636e-14, 2.2442465416964825e-11, -5.27776114586072e-09, 7.610461006178106e-07, -6.148574498547711e-05, 0.00216398089328537]))
    assert v._Tmin_T_trans == v.poly_fit_Tmin
    assert_close1d(v.poly_fit_Tmin_quadratic, [9.044411973585966e-11, 0.0, 3.2636013401752355e-05])

    # furufryl alcohol
    v = VolumeLiquid(poly_fit=(189.6, 568.8000000000001, [6.647049391841594e-25, -1.8278965754470383e-21, 2.170226311490262e-18, -1.4488205117650477e-15, 5.940737543663388e-13, -1.529528764741801e-10, 2.4189938181664235e-08, -2.0920414326130236e-06, 0.00014008105452139704]))
    assert v._Tmin_T_trans == v.poly_fit_Tmin
    assert_close1d(v.poly_fit_Tmin_quadratic, [1.0450567029924047e-10, 2.3832444455827036e-08, 6.0795970965526267e-05])

    # Water - may change, second should always be zero
    v = VolumeLiquid(poly_fit=(273.17, 637.096, [9.00307261049824e-24, -3.097008950027417e-20, 4.608271228765265e-17, -3.8726692841874345e-14, 2.0099220218891486e-11, -6.596204729785676e-09, 1.3368112879131157e-06, -0.00015298762503607717, 0.007589247005014652]))
    assert_close1d(v.poly_fit_Tmin_quadratic,  [8.382589086490995e-12, 0.0, 1.739254665187681e-05])

@pytest.mark.meta_T_dept
def test_VolumeLiquidConstantExtrapolation():
    obj = VolumeLiquid(CASRN="108-38-3", MW=106.165, Tb=412.25, Tc=617.0, Pc=3541000.0,
                       Vc=0.000375, Zc=0.25884384676233363, omega=0.331, dipole=0.299792543559857,
                       Psat=None, extrapolation="constant", method="DIPPR_PERRY_8E")

    assert_close(obj.T_dependent_property(obj.DIPPR_Tmax), 0.0003785956866273838, rtol=1e-9)
    assert_close(obj.T_dependent_property(obj.DIPPR_Tmin), 0.00011563344813684695, rtol=1e-9)


    assert obj.T_dependent_property(618) == obj.T_dependent_property(obj.DIPPR_Tmax)
    assert obj.T_dependent_property(1223532.0) == obj.T_dependent_property(obj.DIPPR_Tmax)

    assert obj.T_dependent_property(1.0) == obj.T_dependent_property(obj.DIPPR_Tmin)
    assert obj.T_dependent_property(.0) == obj.T_dependent_property(obj.DIPPR_Tmin)
    assert obj.T_dependent_property(100) == obj.T_dependent_property(obj.DIPPR_Tmin)



@pytest.mark.meta_T_dept
def test_VolumeLiquidDumpEOS():
    eos = [PR(T=300, P=1E5, Tc=430.8, Pc=7884098.25, omega=0.251)]
    obj = VolumeLiquid(CASRN='7446-09-5', MW=64.0638,  Tc=430.8, Pc=7884098.25, omega=0.251, dipole=1.63, eos=eos)
    hash0 = hash(obj)
    obj2 = VolumeLiquid.from_json(json.loads(json.dumps(obj.as_json())))
    assert obj == obj2
    assert hash(obj) == hash0
    assert hash(obj2) == hash0

    obj = VolumeLiquid(CASRN='7446-09-5', MW=64.0638,  Tc=430.8, Pc=7884098.25, omega=0.251, dipole=1.63)
    hash0 = hash(obj)
    obj2 = VolumeLiquid.from_json(json.loads(json.dumps(obj.as_json())))
    assert obj == obj2
    assert hash(obj) == hash0
    assert hash(obj2) == hash0

@pytest.mark.meta_T_dept
def test_VolumeSolid():
    Vm = VolumeSolid(CASRN='10022-31-8').T_dependent_property(300)
    assert_close(Vm, 8.06592592592592e-05)

    assert None == VolumeSolid(CASRN='10022-31-8').T_dependent_property(-100)

    with pytest.raises(Exception):
        VolumeSolid(CASRN='10022-31-8').test_method_validity(200, 'BADMETHOD')

    BaN2O6 = VolumeSolid(CASRN='10022-31-8')
    BaN2O6.add_tabular_data([200, 300], [8.06e-5, 8.05e-5], 'fake')
    assert_close(8.055e-05, BaN2O6.T_dependent_property(250))
    BaN2O6.tabular_extrapolation_permitted = False
    BaN2O6.test_method_validity(150, 'fake')

@pytest.mark.meta_T_dept
def test_VolumeSolid_works_with_no_data():
    # Test pentane has no property but that it succeeds
    VolumeSolid(CASRN='109-66-0').T_dependent_property(300)

@pytest.mark.fitting
@pytest.mark.meta_T_dept
def test_VolumeSolid_fitting2():
    obj = VolumeSolid(CASRN='7782-44-7', load_data=False)
    Ts_gamma = [42.801, 44.0, 46.0, 48.0,
               50.0, 52.0, 54.0, 54.361]
    
    Vms_gamma = [23.05e-6, 23.06e-6, 23.18e-6, 23.30e-6,
                23.43e-6, 23.55e-6, 23.67e-6, 23.69e-6]
    obj.fit_add_model(Ts=Ts_gamma, data=Vms_gamma, model='DIPPR100', name='gamma')
    assert obj.method == 'gamma'

@pytest.mark.fitting
@pytest.mark.meta_T_dept
def test_VolumeSolid_fitting_wrong_length():
    obj = VolumeSolid(CASRN='7782-44-7', load_data=False)
    Ts_gamma = [42.801, 44.0, 46.0, 48.0,
               50.0, 52.0, 54.0]
    
    Vms_gamma = [23.05e-6, 23.06e-6, 23.18e-6, 23.30e-6,
                23.43e-6, 23.55e-6, 23.67e-6, 23.69e-6]
    with pytest.raises(ValueError):
        obj.fit_add_model(Ts=Ts_gamma, data=Vms_gamma, model='DIPPR100', name='gamma')

@pytest.mark.meta_T_dept
@pytest.mark.fitting
def test_VolumeSolid_fitting1():
    # Initial Result: went poorly + Don't have that exact model
    # 1) A special polynomial fitting may be useful
    # 2) If the model does not "converge", it may be worth trying other solvers
    ammonia_Ts = [194.150, 194.650, 195.150, 195.650, 196.150, 196.650, 197.150, 197.650, 198.150, 198.650, 199.150, 199.650, 200.150, 200.650, 201.150, 201.650, 202.150, 202.650, 203.150, 203.650, 204.150]
    ammonia_Vms = [1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3, 1/47.9730E3]
    obj = VolumeSolid(CASRN='7664-41-7', load_data=False)
    
    res, stats = VolumeSolid.fit_data_to_model(Ts=ammonia_Ts, data=ammonia_Vms, model='DIPPR100',
                          do_statistics=True, use_numba=False)
    assert stats['MAE'] < 1e-8




@pytest.mark.fitting
@pytest.mark.meta_T_dept
def test_VolumeSolid_fitting0():
    Ts_alpha = [4.2, 18.5, 20, 22, 23.880]
    Vms_alpha = [20.75e-6, 20.75e-6, 20.75e-6, 20.78e-6, 20.82e-6]
    
    Ts_beta = [23.880, 24, 26, 28,
               30, 32, 34, 36,
               38, 40, 42, 43.801]
    Vms_beta = [20.95e-6, 20.95e-6, 21.02e-6, 21.08e-6,
                21.16e-6, 21.24e-6, 21.33e-6, 21.42e-6, 
                21.52e-6, 21.63e-6, 21.75e-6, 21.87e-6]
                
    Ts_gamma = [42.801, 44.0, 46.0, 48.0,
               50.0, 52.0, 54.0, 54.361]
    
    Vms_gamma = [23.05e-6, 23.06e-6, 23.18e-6, 23.30e-6,
                23.43e-6, 23.55e-6, 23.67e-6, 23.69e-6]
    
    obj = VolumeSolid(CASRN='7782-44-7', load_data=False)
    # obj.add_tabular_data(Ts=Ts_alpha, properties=Vms_alpha, name='O2_alpha')
    # obj.add_tabular_data(Ts=Ts_beta, properties=Vms_beta, name='O2_beta')
    # obj.add_tabular_data(Ts=Ts_gamma, properties=Vms_gamma, name='O2_gamma')
    
    fit_zeros_specified = obj.fit_data_to_model(Ts=Ts_gamma, data=Vms_gamma, model='DIPPR100', do_statistics=False, use_numba=False,
                     model_kwargs={'C': 0.0, 'D': 0.0, 'E': 0.0, 'F': 0.0, 'G': 0.0, 'B': 0.0})
    for l in ('B', 'C', 'D', 'E', 'F'):
        assert fit_zeros_specified[l] == 0.0
        
    fit_constant = obj.fit_data_to_model(Ts=Ts_gamma, data=Vms_gamma, model='constant', do_statistics=False, use_numba=False)
    assert_close(fit_zeros_specified['A'], fit_constant['A'], rtol=1e-13)

@pytest.mark.meta_T_dept
@pytest.mark.fitting
def test_VolumeLiquid_fitting0():

    ammonia_Ts_V_l = [195, 206.061, 217.121, 228.182, 239.242, 239.82, 250.303, 261.363, 272.424, 283.484, 294.545, 305.605, 316.666, 327.726, 338.787, 349.847, 360.908, 371.968, 383.029, 394.089, 405.15]
    ammonia_V_ls = [1/43200.2, 1/42436.8, 1/41658, 1/40862.4, 1/40048.1, 1/40005, 1/39213, 1/38354.7, 1/37470.3, 1/36556, 1/35607.3, 1/34618.5, 1/33582.2, 1/32488.1, 1/31322.5, 1/30064.9, 1/28683.4, 1/27123.2, 1/25274.3, 1/22840.2, 1/16984.9]
    ammonia_rhom_ls = [1/v for v in ammonia_V_ls]
    obj = VolumeLiquid(CASRN='7664-41-7', load_data=False)

    coeffs, stats = obj.fit_data_to_model(Ts=ammonia_Ts_V_l, data=ammonia_rhom_ls, model='DIPPR105',
                          do_statistics=True, use_numba=False,
                          guesses= {'A': 4.0518E3, 'B': 0.27129, 'C': 405.4,'D': 0.31349})
    assert stats['MAE'] < 1e-5
    
@pytest.mark.meta_T_dept
@pytest.mark.fitting
def test_VolumeLiquid_fitting1_dippr():
    fit_check_CASs = ['124-38-9', '74-98-6', '1333-74-0', '630-08-0', 
                      '100-21-0', '624-92-0', '624-72-6', '74-86-2',
                      '115-07-1', '64-18-6']
    for CAS in fit_check_CASs:
        obj = VolumeLiquid(CASRN=CAS)
        Ts = linspace(obj.DIPPR_Tmin, obj.DIPPR_Tmax, 8)
        props_calc = [1.0/obj.calculate(T, DIPPR_PERRY_8E) for T in Ts]
    
        res, stats = obj.fit_data_to_model(Ts=Ts, data=props_calc, model='DIPPR105',
                              do_statistics=True, use_numba=False, fit_method='lm')
        assert stats['MAE'] < 1e-5

@pytest.mark.slow
@pytest.mark.fuzz
@pytest.mark.fitting
@pytest.mark.meta_T_dept
def test_VolumeLiquid_fitting2_dippr_116_ppds():
    for i, CAS in enumerate(chemicals.volume.rho_data_VDI_PPDS_2.index):
        obj = VolumeLiquid(CASRN=CAS)
        Ts = linspace(obj.T_limits[VDI_PPDS][0], obj.T_limits[VDI_PPDS][1], 8)
        props_calc = [Vm_to_rho(obj.calculate(T, VDI_PPDS), obj.VDI_PPDS_MW) for T in Ts]
        
        res, stats = obj.fit_data_to_model(Ts=Ts, data=props_calc, model='DIPPR116',
                              do_statistics=True, use_numba=False, fit_method='lm', 
                              model_kwargs={'Tc': obj.VDI_PPDS_Tc, 'A': obj.VDI_PPDS_rhoc})
        assert stats['MAE'] < 1e-7

@pytest.mark.meta_T_dept
@pytest.mark.fitting
def test_VolumeLiquid_fitting3():
    # From yaws
    Tc, rhoc, b, n, MW = 627.65, 433.128, 0.233, 0.2587, 66.0
    Ts = linspace(293.15, 298.15, 10)
    props_calc = [Rackett_fit(T, Tc, rhoc, b, n, MW) for T in Ts]
    res, stats = VolumeLiquid.fit_data_to_model(Ts=Ts, data=props_calc, model='Rackett_fit',
                          do_statistics=True, use_numba=False, model_kwargs={'MW':MW, 'Tc': Tc,},
                          fit_method='lm')
    assert stats['MAE'] < 1e-5
    
    # From yaws
    Tc, rhoc, b, n, MW = 1030.0, 1795.0521319999998, 0.96491407, 0.15872, 97.995
    Ts = linspace(315.51, 393.15, 10)
    props_calc = [Rackett_fit(T, Tc, rhoc, b, n, MW) for T in Ts]
    res, stats = VolumeLiquid.fit_data_to_model(Ts=Ts, data=props_calc, model='Rackett_fit',
                          do_statistics=True, use_numba=False, model_kwargs={'MW':MW, 'Tc': Tc,},
                          fit_method='lm')
    assert stats['MAE'] < 1e-5

    # From yaws
    Tc, rhoc, b, n, MW = 545.03, 739.99, 0.3, 0.28571, 105.921
    Ts = linspace(331.15, 332.9, 10)
    props_calc = [Rackett_fit(T, Tc, rhoc, b, n, MW) for T in Ts]
    res, stats = VolumeLiquid.fit_data_to_model(Ts=Ts, data=props_calc, model='Rackett_fit',
                          do_statistics=True, use_numba=False, model_kwargs={'MW':MW, 'Tc': Tc,},
                          fit_method='lm')
    assert stats['MAE'] < 1e-5

    # From yaws
    Tc, rhoc, b, n, MW = 1800.0, 2794.568007, 0.647077183, 0.8, 98.999
    Ts = linspace(703.0, 1620., 10)
    props_calc = [Rackett_fit(T, Tc, rhoc, b, n, MW) for T in Ts]
    res, stats = VolumeLiquid.fit_data_to_model(Ts=Ts, data=props_calc, model='Rackett_fit',
                          do_statistics=True, use_numba=False, model_kwargs={'MW':MW, 'Tc': Tc,},
                          fit_method='lm')
    assert stats['MAE'] < 1e-5


@pytest.mark.meta_T_dept
def test_VolumeLiquidMixture():
#    from thermo.mixture import Mixture
#    m = Mixture(['benzene', 'toluene'], zs=[.5, .5], T=298.15, P=101325.)
    T, P, zs = 298.15, 101325.0, [0.5, 0.5]
    MWs = [78.11184, 92.13842]
    ws = zs_to_ws(zs, MWs=MWs)

    VaporPressures = [
        VaporPressure(CASRN="71-43-2", Tb=353.23, Tc=562.05, Pc=4895000.0, omega=0.212, extrapolation="AntoineAB|DIPPR101_ABC", method=EXP_POLY_FIT, exp_poly_fit=(278.68399999999997, 562.01, [4.547344107145341e-20, -1.3312501882259186e-16, 1.6282983902136683e-13, -1.0498233680158312e-10, 3.535838362096064e-08, -3.6181923213017173e-06, -0.001593607608896686, 0.6373679536454406, -64.4285974110459])),
        VaporPressure(CASRN="108-88-3", Tb=383.75, Tc=591.75, Pc=4108000.0, omega=0.257, extrapolation="AntoineAB|DIPPR101_ABC", method=EXP_POLY_FIT, exp_poly_fit=(178.01, 591.74, [-8.638045111752356e-20, 2.995512203611858e-16, -4.5148088801006036e-13, 3.8761537879200513e-10, -2.0856828984716705e-07, 7.279010846673517e-05, -0.01641020023565049, 2.2758331029405516, -146.04484159879843]))
    ]

    VolumeLiquids = [
        VolumeLiquid(CASRN="71-43-2", MW=78.11184, Tb=353.23, Tc=562.05, Pc=4895000.0, Vc=0.000256, Zc=0.2681535335844513, omega=0.212, dipole=0.0, Psat=VaporPressures[0], extrapolation="constant", method=POLY_FIT, poly_fit=(278.68399999999997, 552.02, [2.5040222732960933e-22, -7.922607445206804e-19, 1.088548130214618e-15, -8.481605391952225e-13, 4.098451788397536e-10, -1.257577461969114e-07, 2.3927976459304723e-05, -0.0025810882828932375, 0.12092854717588034])),
        VolumeLiquid(CASRN="108-88-3", MW=92.13842, Tb=383.75, Tc=591.75, Pc=4108000.0, Vc=0.000316, Zc=0.2638426898300023, omega=0.257, dipole=0.33, Psat=VaporPressures[1], extrapolation="constant", method=POLY_FIT, poly_fit=(178.01, 581.75, [2.2801490297347937e-23, -6.411956871696508e-20, 7.723152902379232e-17, -5.197203733189603e-14, 2.1348482785660093e-11, -5.476649499770259e-09, 8.564670053875876e-07, -7.455178589434267e-05, 0.0028545812080104068]))
    ]

    obj = VolumeLiquidMixture(MWs=MWs, Tcs=[562.05, 591.75], Pcs=[4895000.0, 4108000.0], Vcs=[0.000256, 0.000316], Zcs=[0.2681535335844513, 0.2638426898300023], omegas=[0.212, 0.257],
                              CASs=['71-43-2', '108-88-3'], VolumeLiquids=VolumeLiquids, correct_pressure_pure=False)

    hash0 = hash(obj)
    obj2 = VolumeLiquidMixture.from_json(json.loads(json.dumps(obj.as_json())))
    assert obj == obj2
    assert hash(obj) == hash0
    assert hash(obj2) == hash0

    obj2 = eval(str(obj))
    assert obj == obj2
    assert hash(obj) == hash0
    assert hash(obj2) == hash0


    Vml = 9.858773618507427e-05
    assert_close(obj.calculate(T, P, zs, ws, COSTALD_MIXTURE), Vml, rtol=1e-11)
    obj.method = COSTALD_MIXTURE
    assert_close(obj(T, P, zs, ws),  Vml, rtol=1e-11)

    Vml = 9.815400609778346e-05
    assert_close(obj.calculate(T, P, zs, ws, COSTALD_MIXTURE_FIT), Vml, rtol=1e-11)
    obj.method = COSTALD_MIXTURE_FIT
    assert_close(obj(T, P, zs, ws),  Vml, rtol=1e-11)

    Vml = 9.814047221052766e-05
    assert_close(obj.calculate(T, P, zs, ws, LINEAR), Vml, rtol=1e-11)
    obj.method = LINEAR
    assert_close(obj(T, P, zs, ws),  Vml, rtol=1e-11)

    Vml = 9.7377562180953e-05
    assert_close(obj.calculate(T, P, zs, ws, RACKETT), Vml, rtol=1e-11)
    obj.method = RACKETT
    assert_close(obj(T, P, zs, ws), Vml, rtol=1e-11)

    Vml = 9.810986651973312e-05
    assert_close(obj.calculate(T, P, zs, ws, RACKETT_PARAMETERS), Vml, rtol=1e-11)
    obj.method = RACKETT_PARAMETERS
    assert_close(obj(T, P, zs, ws), Vml, rtol=1e-11)

    # Unhappy paths
    with pytest.raises(Exception):
        obj.calculate(m.T, m.P, m.zs, m.ws, 'BADMETHOD')

    with pytest.raises(Exception):
        obj.test_method_validity(m.T, m.P, m.zs, m.ws, 'BADMETHOD')

@pytest.mark.meta_T_dept
def test_VolumeLiquidMixture_Laliberte():
    # Test Laliberte
#    m = Mixture(['water', 'sulfuric acid'], zs=[0.01, 0.99], T=298.15)
    T = 298.15
    P = 101325.0
    MWs = [18.01528, 98.07848]
    VolumeLiquids = [
        VolumeLiquid(CASRN="7732-18-5", MW=18.01528, Tb=373.124, Tc=647.14, Pc=22048320.0, Vc=5.6000000000000006e-05, Zc=0.22947273972184645, omega=0.344, dipole=1.85, extrapolation="constant", method="POLY_FIT", poly_fit=(273.17, 637.096, [9.00307261049824e-24, -3.097008950027417e-20, 4.608271228765265e-17, -3.8726692841874345e-14, 2.0099220218891486e-11, -6.596204729785676e-09, 1.3368112879131157e-06, -0.00015298762503607717, 0.007589247005014652])),
        VolumeLiquid(CASRN="7664-93-9", MW=98.07848, Tb=610.15, Tc=924.0, Pc=6400000.0, Vc=0.00017769999999999998, Zc=0.14803392201622453, omega=0.494, dipole=2.72, extrapolation="constant", method="POLY_FIT", poly_fit=(277.2, 831.6, [2.2367522080473987e-26, -8.993118194782151e-23, 1.5610918942741286e-19, -1.5237052699290813e-16, 9.134548476347517e-14, -3.438497628955312e-11, 7.949714844467134e-09, -1.0057972242365927e-06, 0.00010045015725585257]))
    ]
    zs = [0.01, 0.99]
    ws = zs_to_ws(zs, MWs=MWs)

    obj = VolumeLiquidMixture(MWs=MWs, CASs=['7732-18-5', '7664-93-9'], VolumeLiquids=VolumeLiquids)
    Vm = obj.mixture_property(T, P, zs, ws)
    assert_close(Vm, 4.824170609370422e-05, rtol=1e-12)


@pytest.mark.meta_T_dept
def test_VolumeLiquidMixture_ExcessVolume():
    # Excess volume
    # Vs -7.5E-7 in ddbst http://www.ddbst.com/en/EED/VE/VE0%20Ethanol%3BWater.php
    T = 298.15
    P = 101325.0
    MWs = [18.01528, 46.06844]
    zs = [1- 0.15600,  0.15600]
    ws = zs_to_ws(zs, MWs)

    VaporPressures = [VaporPressure(CASRN="7732-18-5", Tb=373.124, Tc=647.14, Pc=22048320.0, omega=0.344, extrapolation="AntoineAB|DIPPR101_ABC", method="WAGNER_MCGARRY"),
    VaporPressure(CASRN="64-17-5", Tb=351.39, Tc=514.0, Pc=6137000.0, omega=0.635, extrapolation="AntoineAB|DIPPR101_ABC", method="WAGNER_MCGARRY")]

    drink = VolumeLiquidMixture(MWs=[18.01528, 46.06844], Tcs=[647.14, 514.0], Pcs=[22048320.0, 6137000.0], Vcs=[5.6000000000000006e-05, 0.000168], Zcs=[0.22947273972184645, 0.24125043269792065], omegas=[0.344, 0.635], CASs=['7732-18-5', '64-17-5'], correct_pressure_pure=True, method="LALIBERTE",
        VolumeLiquids=[VolumeLiquid(CASRN="7732-18-5", MW=18.01528, Tb=373.124, Tc=647.14, Pc=22048320.0, Vc=5.6000000000000006e-05, Zc=0.22947273972184645, omega=0.344, dipole=1.85, Psat=VaporPressures[0], extrapolation="constant", method="VDI_PPDS", method_P="COSTALD_COMPRESSED"),
                       VolumeLiquid(CASRN="64-17-5", MW=46.06844, Tb=351.39, Tc=514.0, Pc=6137000.0, Vc=0.000168, Zc=0.24125043269792065, omega=0.635, dipole=1.44, Psat=VaporPressures[1], extrapolation="constant", method="DIPPR_PERRY_8E", method_P="COSTALD_COMPRESSED")]
    )
    V_Ex = drink.excess_property(T, P, zs, ws)
    assert_close(V_Ex, -7.242450496000289e-07, rtol=1e-12)


@pytest.mark.meta_T_dept
def test_VolumeGasMixture():
    zs = [.5, .5]
    T=298.15
    P=1E6
    MWs = [31.9988, 28.0134]
    ws = zs_to_ws(zs, MWs)
    # ['oxygen', 'nitrogen']
    VolumeGases = [VolumeGas(CASRN="7782-44-7", MW=31.9988, Tc=154.58, Pc=5042945.25, omega=0.021, dipole=0.0, eos=[PR(Tc=154.58, Pc=5042945.25, omega=0.021, T=298.15, P=1000000.0)], extrapolation=None, method=None, method_P="EOS"),
                   VolumeGas(CASRN="7727-37-9", MW=28.0134, Tc=126.2, Pc=3394387.5, omega=0.04, dipole=0.0, eos=[PR(Tc=126.2, Pc=3394387.5, omega=0.04, T=298.15, P=1000000.0)], extrapolation=None, method=None, method_P="EOS")]
    eos = PRMIX(T=T, P=P, zs=zs, Tcs=[154.58, 126.2], Pcs=[5042945.25, 3394387.5], omegas=[0.021, 0.04])

    obj = VolumeGasMixture(CASs=["7782-44-7" "7727-37-9"], VolumeGases=VolumeGases, eos=[eos], MWs=MWs)

    assert_close(obj.calculate(T, P, zs, ws, EOS), 0.00246280615920584, rtol=1e-7)
    assert_close(obj.calculate(T, P, zs, ws, IDEAL), 0.002478957029602388, rtol=1e-7)
    assert_close(obj.calculate(T, P, zs, ws, LINEAR), 0.002462690948679184, rtol=1e-7)

    hash0 = hash(obj)
    obj2 = VolumeGasMixture.from_json(json.loads(json.dumps(obj.as_json())))
    assert obj == obj2
    assert hash(obj) == hash0
    assert hash(obj2) == hash0

    obj2 = eval(str(obj))
    assert obj == obj2
    assert hash(obj) == hash0
    assert hash(obj2) == hash0

    # Unhappy paths
    with pytest.raises(Exception):
        obj.calculate(m.T, m.P, m.zs, m.ws, 'BADMETHOD')

    with pytest.raises(Exception):
        obj.test_method_validity(m.T, m.P, m.zs, m.ws, 'BADMETHOD')


@pytest.mark.meta_T_dept
def test_VolumeSupercriticalLiquidMixture():
    eos = PR(Tc=507.6, Pc=3025000, omega=0.2975, T=700., P=1e8)
    obj2 = VolumeSupercriticalLiquid(eos=[eos], Tc=eos.Tc, Pc=eos.Pc, omega=eos.omega)
    V_implemented = obj2.calculate_P(T=700.0, P=1e8, method='EOS')
    assert_close(V_implemented, eos.V_l, rtol=1e-13)
    V_implemented = obj2.calculate_P(T=700.0, P=1e3, method='EOS')
#    assert V_implemented is None

    assert obj2.test_method_validity_P(T=700.0, P=1e8, method='EOS')
    assert not obj2.test_method_validity_P(T=700.0, P=1e4, method='EOS')

@pytest.mark.meta_T_dept
@pytest.mark.CoolProp
@pytest.mark.skipif(not has_CoolProp(), reason='CoolProp is missing')
def test_VolumeSupercriticalLiquidMixtureCoolProp():
    from CoolProp.CoolProp import PropsSI
    obj = VolumeSupercriticalLiquid(CASRN='7732-18-5')
    V_implemented = obj.calculate_P(T=700.0, P=1e8, method='COOLPROP')
    V_CoolProp = 1/PropsSI('DMOLAR', 'T', 700, 'P', 1e8, 'water')
    assert_close(V_implemented, V_CoolProp, rtol=1e-13)
    assert obj.test_method_validity_P(T=700.0, P=1e8, method='COOLPROP')
    assert obj.test_method_validity_P(T=700.0, P=1e4, method='COOLPROP')



# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2020, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

import pytest
from math import isinf, isnan
from fluids.numerics import *
from thermo.flash import *
from thermo.phases import *
from thermo.eos_mix import *
from thermo.eos import *
from thermo.vapor_pressure import VaporPressure
from thermo.heat_capacity import *
from thermo.phase_change import *
from thermo.property_package import StabilityTester


def test_C2_C5_liq_Wilson():
    T, P = 360, 3e6
    # m = Mixture(['ethane', 'n-pentane'], ws=[.5, .5], T=T, P=P)
    zs = [0.7058336794895449, 0.29416632051045505] # m.zs
    HeatCapacityGases = [HeatCapacityGas(poly_fit=(50.0, 1000.0, [7.115386645067898e-21, -3.2034776773408394e-17, 5.957592282542187e-14, -5.91169369931607e-11, 3.391209091071677e-08, -1.158730780040934e-05, 0.002409311277400987, -0.18906638711444712, 37.94602410497228])),
                         HeatCapacityGas(poly_fit=(200.0, 1000.0, [7.537198394065234e-22, -4.946850205122326e-18, 1.4223747507170372e-14, -2.3451318313798008e-11, 2.4271676873997662e-08, -1.6055220805830093e-05, 0.006379734000450042, -1.0360272314628292, 141.84695243411866]))]

    Tcs, Pcs, omegas = [305.32, 469.7], [4872000.0, 3370000.0], [0.098, 0.251]
    eos_kwargs = {'Pcs': Pcs, 'Tcs': Tcs, 'omegas': omegas}
    gas = CEOSGas(PRMIX, eos_kwargs, HeatCapacityGases=HeatCapacityGases, T=T, P=P, zs=zs)
    liq = CEOSLiquid(PRMIX, eos_kwargs, HeatCapacityGases=HeatCapacityGases, T=T, P=P, zs=zs)
    stab = StabilityTester(Tcs=Tcs, Pcs=Pcs, omegas=omegas)

    assert_stab_success_2P(liq, gas, stab, 360.0, 3e6, zs, 'Wilson liquid', xs=[0.8178671482462099, 0.18213285175379013],
                       ys=[0.3676551321096354, 0.6323448678903647], VF=0.24884602085493626)


    # 30 Pa off from a bubble point near the critical point
    assert_stab_success_2P(liq, gas, stab, 123.3+273.15, 5475679.470049857, zs, 'Wilson liquid', xs=[0.7058338708478485, 0.29416612915215135],
                           ys=[0.4951812815325094, 0.5048187184674905], VF=9.084070806296505e-07, SS_tol=1e-25, rtol=1e-6)

    # 0.1 Pa off - very stupid/bad/evil, VF cannot be pinned down to ~ a fourth digit
    assert_stab_success_2P(liq, gas, stab, 123.3+273.15, 5475649.470049857+.01, zs, 'Wilson liquid', xs=[0.7058336795533356, 0.2941663204466643],
                           ys=[0.49517801297865455, 0.5048219870213455], VF=3.0282003822009134e-10, SS_tol=1e-25, rtol=1e-3)

    # Check maxiter > 36 allowed for stab convergence; check gas converges
    assert_stab_success_2P(liq, gas, stab, 383, 6e6, zs, 'Wilson gas', xs=[0.6068839791378426, 0.3931160208621572],
                           ys=[0.7735308634810933, 0.22646913651890652], VF=0.5937688162366092, rtol=5e-6)


def test_guesses_bad_ranges():
    stab = StabilityTester(Tcs=[647.086, 514.7], Pcs=[22048320.0, 6137000.0], omegas=[0.344, 0.635], aqueous_check=True, CASs=['7732-18-5', '64-17-5'])
    guesses = list(stab.incipient_guesses(T=6.2, P=5e4, zs=[.4, .6]))
    assert all(not isinf(x) and not isnan(x) for r in guesses for x in r)